# lsqfitgp/_utils.py
#
# Copyright (c) 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import textwrap

def append_to_docstring(docs, doctail, front=False):
    doctail = textwrap.dedent(doctail)
    dedocs = textwrap.dedent(docs)
    lineend = docs.find('\n')
    indented_lineend = dedocs.find('\n')
    indent = docs[:indented_lineend - lineend]
    if front:
        newdocs = doctail + dedocs
    else:
        newdocs = dedocs + doctail
    return textwrap.indent(newdocs, indent)

def top_bottom_rule(title, body):
    body = textwrap.dedent(body)
    body_lines = list(map(str.rstrip, body.split('\n')))
    body = '\n'.join(body_lines)
    body_width = max(map(len, body_lines))
    title_width = 2 + len(title)
    width = max(body_width, title_width + 4)
    pre_length = (width - title_width) // 2
    post_length = width - title_width - pre_length
    toprule = '=' * pre_length + ' ' + title + ' ' + '=' * post_length
    bottomrule = '=' * width
    return '\n'.join([toprule, body, bottomrule])



# lsqfitgp/__init__.py
#
# Copyright (c) 2020, 2022, 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""
Module to fit Gaussian processes

See the manual at https://gattocrucco.github.io/lsqfitgp/docs
"""

__version__ = '0.21'

# these first because they modify global state
from . import _patch_jax
from . import _patch_gvar

from ._array import (
    StructuredArray,
    broadcast,
    broadcast_to,
    broadcast_arrays,
    asarray,
    unstructured_to_structured,
)
from ._Deriv import Deriv
from ._Kernel import (
    CrossKernel,
    Kernel,
    CrossStationaryKernel,
    StationaryKernel,
    CrossIsotropicKernel,
    IsotropicKernel,
    crosskernel,
    kernel,
    crossstationarykernel,
    stationarykernel,
    crossisotropickernel,
    isotropickernel,
)
from ._kernels import * # safe, _kernels/__init__.py only imports kernels
from ._gvarext import (
    jacobian,
    from_jacobian,
    gvar_gufunc,
    switchgvar,
    uformat,
    fmtspec_kwargs,
    gvar_format,
)
from ._GP import GP
from ._fit import empbayes_fit
from ._fastraniter import raniter, sample

from . import copula
from . import bayestree



# lsqfitgp/copula/_beta.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" JAX-compatible implementation of the beta distribution """

import functools

from scipy import special
import jax
from jax.scipy import special as jspecial
from jax import numpy as jnp

from .. import _jaxext

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1))
def betaincinv(a, b, y):
    a = jnp.asarray(a)
    b = jnp.asarray(b)
    y = jnp.asarray(y)
    dtype = _jaxext.float_type(a.dtype, b.dtype, y.dtype)
    return _jaxext.pure_callback_ufunc(
        lambda *args: special.betaincinv(*args).astype(dtype),
        dtype, a, b, y,
    )

dIdx_ = _jaxext.elementwise_grad(jspecial.betainc, 2)

@betaincinv.defjvp
def betaincinv_jvp(a, b, primals, tangents):
    y, = primals
    yt, = tangents
    x = betaincinv(a, b, y)
    dIdx = dIdx_(a, b, x)
    return x, yt / dIdx

class beta:
    
    @staticmethod
    def ppf(q, a, b):
        return betaincinv(a, b, q)



# lsqfitgp/bayestree/_bart.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import numpy
from jax import numpy as jnp
import jax
import gvar

from .. import copula
from .. import _kernels
from .. import _fit
from .. import _array
from .. import _GP
from .. import _fastraniter

# TODO I added a lot of functionality to bcf. The easiest way to port it over is
# adding the option in bcf to drop the second bart model and its associated
# hypers, and then write bart as a simple convenience wrapper-subclass over bcf.

class bart:
    
    def __init__(self,
        x_train,
        y_train,
        *,
        weights=None,
        fitkw={},
        kernelkw={},
        marginalize_mean=True,
    ):
        """
        Nonparametric Bayesian regression with a GP version of BART.

        Evaluate a Gaussian process regression with a kernel which accurately
        approximates the infinite trees limit of BART. The hyperparameters are
        optimized to their marginal MAP.

        Parameters
        ----------
        x_train : (n, p) array or dataframe
            Observed covariates.
        y_train : (n,) array
            Observed outcomes.
        weights : (n,) array
            Weights used to rescale the error variance (as 1 / weight).
        fitkw : dict
            Additional arguments passed to `~lsqfitgp.empbayes_fit`, overrides
            the defaults.
        kernelkw : dict
            Additional arguments passed to `~lsqfitgp.BART`, overrides the
            defaults.
        marginalize_mean : bool
            If True (default), marginalize the intercept of the model.
        
        Notes
        -----
        The regression model is:

        .. math::
            y_i &= \\mu + \\lambda f(\\mathbf x_i) + \\varepsilon_i, \\\\
            \\varepsilon_i &\\overset{\\mathrm{i.i.d.}}{\\sim}
                N(0, \\sigma^2 / w_i), \\\\
            \\mu &\\sim N(
                (\\max(\\mathbf y) + \\min(\\mathbf y)) / 2,
                (\\max(\\mathbf y) - \\min(\\mathbf y))^2 / 4
            ), \\\\
            \\log \\sigma^2 &\\sim N(
                \\log(\\overline{w(y - \\bar y)^2}),
                4
            ), \\\\
            \\log \\lambda &\\sim N(
                \\log ((\\max(\\mathbf y) - \\min(\\mathbf y)) / 4),
                4
            ), \\\\
            f &\\sim \\mathrm{GP}(
                0,
                \\mathrm{BART}(\\alpha,\\beta)
            ), \\\\
            \\alpha &\\sim \\mathrm{B}(2, 1), \\\\
            \\beta &\\sim \\mathrm{IG}(1, 1).

        To make the inference, :math:`(f, \\boldsymbol\\varepsilon, \\mu)` are
        marginalized analytically, and the marginal posterior mode of
        :math:`(\\sigma, \\lambda, \\alpha, \\beta)` is found by numerical
        minimization, after transforming them to express their prior as a
        Gaussian copula. Their marginal posterior covariance matrix is estimated
        with an approximation of the hessian inverse. See
        `~lsqfitgp.empbayes_fit` and use the parameter ``fitkw`` to customize
        this procedure.

        The tree splitting grid of the BART kernel is set using quantiles of the
        observed covariates. This corresponds to settings ``usequants=True``,
        ``numcut=inf`` in the R packages BayesTree and BART. Use the
        ``kernelkw`` parameter to customize the grid.

        Attributes
        ----------
        mean : gvar
            The prior mean :math:`\\mu`.
        sigma : float or gvar
            The error term standard deviation :math:`\\sigma`. If there are
            weights, the sdev for each unit is obtained dividing ``sigma`` by
            sqrt(weight).
        alpha : gvar
            The numerator of the tree spawn probability :math:`\\alpha` (named
            ``base`` in BayesTree and BART).
        beta : gvar
            The depth exponent of the tree spawn probability :math:`\\beta`
            (named ``power`` in BayesTree and BART).
        meansdev : gvar
            The prior standard deviation :math:`\\lambda` of the latent
            regression function.
        fit : empbayes_fit
            The hyperparameters fit object.

        Methods
        -------
        gp :
            Create a GP object.
        data :
            Creates the dictionary to be passed to `GP.pred` to represent
            ``y_train``.
        pred :
            Evaluate the regression function at given locations.

        See also
        --------
        lsqfitgp.BART
        
        """

        # convert covariates to StructuredArray
        x_train = self._to_structured(x_train)
    
        # convert outcomes to 1d array
        if hasattr(y_train, 'to_numpy'):
            y_train = y_train.to_numpy()
            y_train = y_train.squeeze() # for dataframes
        y_train = jnp.asarray(y_train)
        assert y_train.shape == x_train.shape

        # check weights
        self._no_weights = weights is None
        if self._no_weights:
            weights = jnp.ones_like(y_train)
        assert weights.shape == y_train.shape
    
        # prior mean and variance
        ymin = jnp.min(y_train)
        ymax = jnp.max(y_train)
        mu_mu = (ymax + ymin) / 2
        k_sigma_mu = (ymax - ymin) / 2
        
        # splitting points and indices
        splits = _kernels.BART.splits_from_coord(x_train)
        i_train = self._toindices(x_train, splits)

        # prior on hyperparams
        sigma2_priormean = numpy.mean((y_train - y_train.mean()) ** 2 * weights)
        hyperprior = copula.makedict({
            'alpha': copula.beta(2, 1), # base of tree gen prob
            'beta': copula.invgamma(1, 1), # exponent of tree gen prob
            'log(k)': gvar.gvar(numpy.log(2), 2), # denominator of prior sdev
            'log(sigma2)': gvar.gvar(numpy.log(sigma2_priormean), 2),
                # i.i.d. error variance, scaled with weights
            'mean': gvar.gvar(mu_mu, k_sigma_mu), # mean of the GP
        })
        if marginalize_mean:
            hyperprior.pop('mean')

        # GP factory
        def makegp(hp, *, i_train, weights, splits, **_):
            kw = dict(
                alpha=hp['alpha'], beta=hp['beta'],
                maxd=10, reset=[2, 4, 6, 8], gamma=0.95,
            )
            kw.update(kernelkw)
            kernel = _kernels.BART(splits=splits, indices=True, **kw)
            kernel *= (k_sigma_mu / hp['k']) ** 2
            
            gp = (_GP
                .GP(kernel, checkpos=False, checksym=False, solver='chol')
                .addx(i_train, 'trainmean')
                .addcov(jnp.diag(hp['sigma2'] / weights), 'trainnoise')
            )
            pieces = {'trainmean': 1, 'trainnoise': 1}
            if 'mean' not in hp:
                gp = gp.addcov(k_sigma_mu ** 2, 'mean')
                pieces.update({'mean': 1})
            return gp.addtransf(pieces, 'train')
            
        # data factory
        def info(hp, *, mu_mu, **_):
            return {'train': y_train - hp.get('mean', mu_mu)}

        # fit hyperparameters
        gpkw = dict(
            i_train=i_train,
            weights=weights,
            splits=splits,
            mu_mu=mu_mu,
        )
        options = dict(
            verbosity=3,
            raises=False,
            minkw=dict(method='l-bfgs-b', options=dict(maxls=4, maxiter=100)),
            mlkw=dict(epsrel=0),
            forward=True,
            gpfactorykw=gpkw,
        )
        options.update(fitkw)
        fit = _fit.empbayes_fit(hyperprior, makegp, info, **options)
        
        # extract hyperparameters from minimization result
        self.sigma = gvar.sqrt(fit.p['sigma2'])
        self.alpha = fit.p['alpha']
        self.beta = fit.p['beta']
        self.meansdev = k_sigma_mu / fit.p['k']
        self.mean = fit.p.get('mean', mu_mu)

        # set public attributes
        self.fit = fit

        # set private attributes
        self._ystd = y_train.std()

    def _gethp(self, hp, rng):
        if not isinstance(hp, str):
            return hp
        elif hp == 'map':
            return self.fit.pmean
        elif hp == 'sample':
            return _fastraniter.sample(self.fit.pmean, self.fit.pcov, rng=rng)
        else:
            raise KeyError(hp)

    def gp(self, *, hp='map', x_test=None, weights=None, rng=None):
        """
        Create a Gaussian process with the fitted hyperparameters.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'``, use the marginal maximum a
            posteriori. If ``'sample'``, sample hyperparameters from the
            posterior. If a dict, use the given hyperparameters.
        x_test : array or dataframe, optional
            Additional covariates for "test points".
        weights : array, optional
            Weights for the error variance on the test points.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        gp : GP
            A centered Gaussian process object. To add the mean, use the
            ``mean`` attribute of the `bart` object. The keys of the GP are
            'Xmean', 'Xnoise', and 'X', where the "X" stands either for 'train'
            or 'test', and X = Xmean + Xnoise.
        """

        hp = self._gethp(hp, rng)
        return self._gp(hp, x_test, weights, self.fit.gpfactorykw)

    def _gp(self, hp, x_test, weights, gpfactorykw):

        # create GP object
        gp = self.fit.gpfactory(hp, **gpfactorykw)

        # add test points
        if x_test is not None:

            # convert covariates to indices
            x_test = self._to_structured(x_test)
            i_test = self._toindices(x_test, gpfactorykw['splits'])
            assert i_test.dtype == gpfactorykw['i_train'].dtype

            # check weights
            if weights is not None:
                weights = jnp.asarray(weights)
                assert weights.shape == i_test.shape
            else:
                weights = jnp.ones(i_test.shape)

            # add test points
            gp = (gp
                .addx(i_test, 'testmean')
                .addcov(jnp.diag(hp['sigma2'] / weights), 'testnoise')
            )
            pieces = {'testmean': 1, 'testnoise': 1}
            if 'mean' not in hp:
                pieces.update({'mean': 1})
            gp = gp.addtransf(pieces, 'test')

        return gp

    def data(self, *, hp='map', rng=None):
        """
        Get the data to be passed to `GP.pred` on a GP object returned by `gp`.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'``, use the marginal maximum a
            posteriori. If ``'sample'``, sample hyperparameters from the
            posterior. If a dict, use the given hyperparameters.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        data : dict
            A dictionary representing ``y_train`` in the format required by the
            `GP.pred` method.
        """

        hp = self._gethp(hp, rng)
        return self.fit.data(hp, **self.fit.gpfactorykw)

    def pred(self, *, hp='map', error=False, format='matrices', x_test=None,
        weights=None, rng=None):
        """
        Predict the outcome at given locations.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'``, use the marginal maximum a
            posteriori. If ``'sample'``, sample hyperparameters from the
            posterior. If a dict, use the given hyperparameters.
        error : bool
            If ``False`` (default), make a prediction for the latent mean. If
            ``True``, add the error term.     
        format : {'matrices', 'gvar'}
            If 'matrices' (default), return the mean and covariance matrix
            separately. If 'gvar', return an array of gvars.
        x_test : array or dataframe, optional
            Covariates for the locations where the prediction is computed. If
            not specified, predict at the data covariates.
        weights : array, optional
            Weights for the error variance on the test points.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        If ``format`` is 'matrices' (default):

        mean, cov : arrays
            The mean and covariance matrix of the Normal posterior distribution
            over the regression function at the specified locations.

        If ``format`` is 'gvar':

        out : array of `GVar`
            The same distribution represented as an array of `GVar` objects.
        """

        # TODO it is a bit confusing that if x_test=None and error=True, the
        # prediction returns y_train exactly, instead of hypothetical new
        # observations at the same covariates.
        
        hp = self._gethp(hp, rng)
        if x_test is not None:
            x_test = self._to_structured(x_test)
        mean, cov = self._pred(hp, x_test, weights, self.fit.gpfactorykw, bool(error))

        if format == 'gvar':
            return gvar.gvar(mean, cov, fast=True)
        elif format == 'matrices':
            return mean, cov
        else:
            raise KeyError(format)

    @functools.cached_property
    def _pred(self):
        
        @functools.partial(jax.jit, static_argnums=(4,))
        def _pred(hp, x_test, weights, gpfactorykw, error):
            gp = self._gp(hp, x_test, weights, gpfactorykw)
            data = self.fit.data(hp, **gpfactorykw)
            if x_test is None:
                label = 'train'
            else:
                label = 'test'
            if not error:
                label += 'mean'
            outmean, outcov = gp.predfromdata(data, label, raw=True)
            return outmean + hp.get('mean', gpfactorykw['mu_mu']), outcov

        return _pred

    @classmethod
    def _to_structured(cls, x):

        # convert to StructuredArray
        if hasattr(x, 'columns'):
            x = _array.StructuredArray.from_dataframe(x)
        elif x.dtype.names is None:
            x = _array.unstructured_to_structured(x)
        else:
            x = _array.StructuredArray(x)

        # check
        assert x.ndim == 1
        def check_numerical(path, dtype):
            if not numpy.issubdtype(dtype, numpy.number):
                raise TypeError(f'covariate `{path}` is not numerical')
        cls._walk_dtype(x.dtype, check_numerical)

        return x

    @classmethod
    def _walk_dtype(cls, dtype, task, path=None):
        if dtype.names is None:
            task(path, dtype)
        else:
            for name in dtype.names:
                subpath = name if path is None else path + ':' + name
                cls._walk_dtype(dtype[name], task, subpath)

    @staticmethod
    def _toindices(x, splits):
        ix = _kernels.BART.indices_from_coord(x, splits)
        return _array.unstructured_to_structured(ix, names=x.dtype.names)

    def __repr__(self):
        out = f"""BART fit:
alpha = {self.alpha} (0 -> intercept only, 1 -> any)
beta = {self.beta} (0 -> any, ∞ -> no interactions)
mean = {self.mean}
latent sdev = {self.meansdev} (large -> conservative extrapolation)
data total sdev = {self._ystd:.3g}"""

        if self._no_weights:
            out += f"""
error sdev = {self.sigma}"""
        else:
            weights = numpy.array(self.fit.gpfactorykw['weights'])
            avgsigma = numpy.sqrt(numpy.mean(self.sigma ** 2 / weights))
            out += f"""
error sdev (avg weighted) = {avgsigma}
error sdev (unweighted) = {self.sigma}"""

        return out



# lsqfitgp/_linalg/_stdcplx.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""
Module to estimate the time taken by standard linear algebra operations.
"""

# TODO maybe I can replace this with jax compilation introspection, which
# provides flops estimates. See https://jax.readthedocs.io/en/latest/aot.html

import timeit
import inspect

from jax import random
from jax import numpy as jnp
from jax.scipy import linalg as jlinalg
from scipy import sparse

def benchmark(func, *args, **kwargs):
    timer = timeit.Timer('func(*args, **kwargs)', globals=locals())
    n, _ = timer.autorange()
    times = timer.repeat(5, n)
    time = min(times) / n
    return time

ops = {
    'chol': [
        lambda x: jnp.linalg.cholesky(x), # function performing the operation
        lambda s: s[0] ** 3, # complexity in terms of arguments' shapes
    ],
    'eigh': [
        lambda x: jnp.linalg.eigh(x),
        lambda s: s[0] ** 3,
    ],
    'qr-red': [
        lambda x: jnp.linalg.qr(x, mode='reduced'),
        lambda s: min(s) ** 2 * max(s),
    ],
    'qr-full': [
        lambda x: jnp.linalg.qr(x, mode='full'),
        lambda s: max(s) ** 3,
    ],
    'svd-red': [
        lambda x: jnp.linalg.svd(x, full_matrices=False),
        lambda s: min(s) ** 2 * max(s),
    ],
    'svd-full': [
        lambda x: jnp.linalg.svd(x, full_matrices=True),
        lambda s: max(s) ** 3,
    ],
    'solve_triangular': [
        lambda x, y: jlinalg.solve_triangular(x, y),
        lambda s, t: s[0] ** 2 * t[1],
    ],
    'matmul': [
        lambda x, y: jnp.matmul(x, y),
        lambda s, t: s[0] * s[1] * t[1],
    ]
}

def gen_ops_factors(n): # pragma: no cover
    key = random.PRNGKey(202208101236)
    factors = {}
    for op, (job, est) in ops.items():
        print(f'{op}({n})... ', end='', flush=True)
        nparams = len(inspect.signature(job).parameters)
        key, subkey = random.split(key)
        m = random.normal(subkey, (nparams, n, n), jnp.float32)
        args = m @ jnp.swapaxes(m, -2, -1)
        time = benchmark(job, *args)
        print(f'{time:.2g} s')
        factors[op] = time / est(*(a.shape for a in args))
    return factors

ops_factors = {'chol': 6.03470915928483e-12,
 'eigh': 1.824986875290051e-10,
 'qr-red': 1.1241237493231893e-10,
 'qr-full': 1.2058762495871633e-10,
 'svd-red': 4.468000000342727e-10,
 'svd-full': 4.2561762500554324e-10,
 'solve_triangular': 4.1634716559201486e-12,
 'matmul': 5.6301691802218555e-12} # = gen_ops_factors(1000)

ops_consts = {'chol': 1.810961455339566e-06,
 'eigh': 2.390482500195503e-06,
 'qr-red': 2.6676162518560884e-06,
 'qr-full': 2.6932845800183714e-06,
 'svd-red': 3.7152979196980598e-06,
 'svd-full': 3.663789590355009e-06,
 'solve_triangular': 2.170706249307841e-06,
 'matmul': 1.718031665077433e-06} # = gen_ops_factors(1)

def predtime(op, shapes, types):
    """
    Estimate the time taken by a linear algebra operation.
    
    Parameters
    ----------
    op : str
        The identifier of the operation, see `listops`.
    shapes : sequence of tuples of integers
        The shapes of the arguments.
    types : sequence of numpy data types
        The types of the arguments. They are promoted according to JAX rules.
    
    Returns
    -------
    time : float
        The estimated time. Not accurate. The unit of measure is seconds on a
        particular laptop cpu used to calibrate the estimate with 1000x1000
        matrices.
    """
    _, est = ops[op]
    factor = ops_factors[op]
    const = ops_consts[op]
    dt = jnp.sin(jnp.empty(0, jnp.result_type(*types))).dtype
    if dt == jnp.float64:
        factor *= 2
    return const + factor * est(*shapes)

def listops():
    """
    List available linear algebra operations.
    
    Returns
    -------
    ops : dict
        A dictionary operation identifier -> number of arguments.
    """
    return {
        op: len(inspect.signature(job).parameters)
        for op, (job, _) in ops.items()
    }

# TODO I should estimate the cost under jit, and subtract the overheaded
# estimated with a jitted no-op.



# lsqfitgp/_special/__init__.py
#
# Copyright (c) 2022, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from ._bernoulli import (
    periodic_bernoulli,
    scaled_periodic_bernoulli,
)
from ._bessel import (
    j0,
    j1,
    jv,
    jvp,
    kv,
    kvp,
    iv,
    ivp,
    jvmodx2,
    kvmodx2,
    kvmodx2_hi,
)
from ._exp import (
    expm1x,
)
from ._expint import (
    expn_imag,
    exp1_imag,
    ci,
)
from ._gamma import (
    sgngamma,
    gamma,
    poch,
    gamma_incr,
    gammaln1,
)
from ._sinc import sinc
from ._taylor import taylor
from ._zeta import (
    zeta,
    hurwitz_zeta,
    periodic_zeta,
    zeta_zero,
    zeta_series_power_diff,
)



# lsqfitgp/_Kernel/_decorators.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import types
import warnings
import inspect

from . import _crosskernel
from . import _kernel
from . import _stationary
from . import _isotropic

def makekernelsubclass(core, bases, **prekw):

    named_object = getattr(core, 'pyfunc', core) # np.vectorize objects
    name = getattr(named_object, '__name__', 'DecoratedKernel')

    bases = tuple(bases)

    def exec_body(ns):
        
        def __new__(cls, **kw):
            kwargs = prekw.copy()
            kwargs.update(kw)
            if len(kwargs) < len(prekw) + len(kw):
                shared_keys = set(prekw).intersection(kw)
                warnings.warn(f'overriding init argument(s) '
                    f'{shared_keys} of kernel {name}')
            self = super(newclass, cls).__new__(cls, core, **kwargs)
            if isinstance(self, bases[-1]) and set(kw).issubset(self.initkw):
                self = self._clone(cls)
            return self
        
        ns['__new__'] = __new__
        ns['__wrapped__'] = named_object
        ns['__doc__'] = named_object.__doc__

    newclass = types.new_class(name, bases, exec_body=exec_body)
    assert issubclass(newclass, _crosskernel.CrossKernel)
    return newclass

def crosskernel(*args, bases=None, **kw):
    """
    
    Decorator to convert a function to a subclass of `CrossKernel`.

    Parameters
    ----------
    *args :
        Either a function to decorate, or no arguments. The function is used
        as the `core` argument to `CrossKernel`.
    bases : tuple of types, optional
        The bases of the new class. If not specified, use `CrossKernel`.
    **kw :
        Additional arguments are passed to `CrossKernel`.

    Returns
    -------
    class_or_dec : callable or type
        If `args` is empty, a decorator ready to be applied, else the kernel
        class.

    Examples
    --------
    
    >>> @lgp.crosskernel(derivable=True)
    ... def MyKernel(x, y, a=0, b=0):
    ...     return (x - a) * (y - b)

    Notes
    -----
    Arguments passed to the class constructor may modify the class. If the
    object returned by the the constructor is a subclass of the superclass
    targeted by the decorator, and all the arguments passed at instantiation
    are passed down to the decorated function, the class of the object is
    enforced to be the new class.
    
    """
    if bases is None:
        bases = _crosskernel.CrossKernel,
    functional = lambda core: makekernelsubclass(core, bases, **kw)
    if len(args) == 0:
        return functional
    elif len(args) == 1:
        return functional(*args)
    else:
        raise ValueError(len(args))

def kernel(*args, **kw):
    """
    
    Like `crosskernel` but makes a subclass of `Kernel`.

    Examples
    --------
    
    >>> @lgp.kernel(loc=10) # the default loc will be 10
    ... def MyKernel(x, y, cippa=1, lippa=42):
    ...     return cippa * (x * y) ** lippa
    
    """
    return crosskernel(*args, bases=(_kernel.Kernel,), **kw)

def crossstationarykernel(*args, **kw):
    """
    
    Like `crosskernel` but makes a subclass of `CrossStationaryKernel`.
    
    """
    return crosskernel(*args, bases=(_stationary.CrossStationaryKernel,), **kw)

def stationarykernel(*args, **kw):
    """
    
    Like `crosskernel` but makes a subclass of `StationaryKernel`.

    Examples
    --------
    
    >>> @lgp.stationarykernel(input='posabs')
    ... def MyKernel(absdelta, cippa=1, lippa=42):
    ...     return cippa * sum(
    ...         jnp.exp(-absdelta[name] / lippa)
    ...         for name in absdelta.dtype.names
    ...     )
    
    """
    return crosskernel(*args, bases=(_stationary.StationaryKernel,), **kw)

def crossisotropickernel(*args, **kw):
    """
    
    Like `crosskernel` but makes a subclass of `CrossIsotropicKernel`.
    
    """
    return crosskernel(*args, bases=(_isotropic.CrossIsotropicKernel,), **kw)

def isotropickernel(*args, **kw):
    """
    
    Like `crosskernel` but makes a subclass of `IsotropicKernel`.

    Examples
    --------
    
    >>> @lgp.isotropickernel(derivable=True)
    ... def MyKernel(distsquared, cippa=1, lippa=42):
    ...     return cippa * jnp.exp(-distsquared) + lippa
    
    """
    return crosskernel(*args, bases=(_isotropic.IsotropicKernel,), **kw)



# lsqfitgp/_kernels/_celerite.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import jax
from jax import numpy as jnp

from .. import _jaxext
from .._Kernel import stationarykernel
    
def _Celerite_derivable(**kw):
    gamma = kw.get('gamma', 1)
    B = kw.get('B', 0)
    if jnp.isscalar(gamma) and jnp.isscalar(B) and B == gamma:
        return 1
    else:
        return False

@stationarykernel(derivable=_Celerite_derivable, input='abs', maxdim=1)
def Celerite(delta, gamma=1, B=0):
    """
    Celerite kernel.
    
    .. math::
        k(\\Delta) = \\exp(-\\gamma|\\Delta|)
        \\big( \\cos(\\Delta) + B \\sin(|\\Delta|) \\big)
    
    This is the covariance function of an AR(2) process with complex roots. The
    parameters must satisfy the condition :math:`|B| \\le \\gamma`. For
    :math:`B = \\gamma` it is equivalent to the `Harmonic` kernel with
    :math:`\\eta Q = 1/B, Q > 1`, and it is derivable.
    
    Reference: Daniel Foreman-Mackey, Eric Agol, Sivaram Ambikasaran, and Ruth
    Angus: *Fast and Scalable Gaussian Process Modeling With Applications To
    Astronomical Time Series*.
    """
    with _jaxext.skipifabstract():
        assert 0 <= gamma < jnp.inf, gamma
        assert abs(B) <= gamma, (B, gamma)
    return jnp.exp(-gamma * delta) * (jnp.cos(delta) + B * jnp.sin(delta))

@stationarykernel(derivable=1, maxdim=1)
def Harmonic(delta, Q=1):
    """
    Damped stochastically driven harmonic oscillator kernel.
    
    .. math::
        k(\\Delta) =
        \\exp\\left( -\\frac {|\\Delta|} {Q} \\right)
        \\begin{cases}
            \\cosh(\\eta\\Delta) + \\sinh(\\eta|\\Delta|) / (\\eta Q)
            & 0 < Q < 1 \\\\
            1 + |\\Delta| & Q = 1 \\\\
            \\cos(\\eta\\Delta) + \\sin(\\eta|\\Delta|) / (\\eta Q)
            & Q > 1,
        \\end{cases}
    
    where :math:`\\eta = \\sqrt{|1 - 1/Q^2|}`.
    
    The process is the solution to the stochastic differential equation
    
    .. math::   f''(x) + 2/Q f'(x) + f(x) = w(x),
    
    where :math:`w` is white noise.
    
    The parameter :math:`Q` is the quality factor, i.e., the ratio between the energy
    stored in the oscillator and the energy lost in each cycle due to damping.
    The angular frequency is 1, i.e., the period is 2π. The process is derivable
    one time.
    
    In 1D, for :math:`Q = 1` (default) and ``scale=sqrt(1/3)``, it is the Matérn 3/2
    kernel.
    
    Reference: Daniel Foreman-Mackey, Eric Agol, Sivaram Ambikasaran, and Ruth
    Angus: *Fast and Scalable Gaussian Process Modeling With Applications To
    Astronomical Time Series*.
    """
    
    # TODO improve and test the numerical accuracy for derivatives near x=0
    # and Q=1. I don't know if the derivatives have problems away from Q=1.
    
    # TODO probably second derivatives w.r.t. Q at Q=1 are wrong.
    
    # TODO will fail if Q is traced.
    
    with _jaxext.skipifabstract():
        assert 0 < Q < jnp.inf, Q
    
    tau = jnp.abs(delta)
    
    if Q < 1/2:
        etaQ = jnp.sqrt((1 - Q) * (1 + Q))
        tauQ = tau / Q
        pexp = jnp.exp(_sqrt1pm1(-jnp.square(Q)) * tauQ)
        mexp = jnp.exp(-(1 + etaQ) * tauQ)
        return (pexp + mexp + (pexp - mexp) / etaQ) / 2
    
    elif 1/2 <= Q < 1:
        etaQ = jnp.sqrt(1 - jnp.square(Q))
        tauQ = tau / Q
        etatau = etaQ * tauQ
        return jnp.exp(-tauQ) * (jnp.cosh(etatau) + jnp.sinh(etatau) / etaQ)
        
    elif Q == 1:
        return _harmonic(tau, Q)
    
    else: # Q > 1
        etaQ = jnp.sqrt(jnp.square(Q) - 1)
        tauQ = tau / Q
        etatau = etaQ * tauQ
        return jnp.exp(-tauQ) * (jnp.cos(etatau) + jnp.sin(etatau) / etaQ)

def _sqrt1pm1(x):
    """sqrt(1 + x) - 1, numerically stable for small x"""
    return jnp.expm1(1/2 * jnp.log1p(x))

@jax.custom_jvp
def _matern32(x):
    return (1 + x) * jnp.exp(-x)

_matern32.defjvps(lambda g, ans, x: g * -x * jnp.exp(-x))

def _harmonic(x, Q):
    return _matern32(x / Q) + jnp.exp(-x/Q) * (1 - Q) * jnp.square(x) * (1 + x/3)

# def _harmonic(x, Q):
#     return np.exp(-x/Q) * (1 + x + (1 - Q) * x * (1 + x * (1 + x/3)))

# @autograd.extend.primitive
# def _harmonic(x, Q):
#     return (1 + x) * np.exp(-x)
#
# autograd.extend.defvjp(
#     _harmonic,
#     lambda ans, x, Q: lambda g: g * -np.exp(-x/Q) * x * (1 + (Q-1) * (1+x)),
#     lambda ans, x, Q: lambda g: g * -np.exp(-x) * x ** 3 / 3
# ) # d/dQ: -np.exp(-x/Q) * (3/Q**2 - 1) * x**3 / (6 * Q**2)
#
# autograd.extend.defjvp(
#     _harmonic,
#     lambda g, ans, x, Q: (g.T * (-np.exp(-x) * x).T).T,
#     lambda g, ans, x, Q: (g.T * (-np.exp(-x) * x ** 3 / 3).T).T
# )



# lsqfitgp/_gvarext/_jacobian.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import gvar
import numpy

def _getsvec(x):
    """
    Get the sparse vector of derivatives of a GVar.
    """
    if isinstance(x, gvar.GVar):
        return x.internaldata[1]
    else:
        return gvar.svec(0)

def _merge_svec(gvlist, start=None, stop=None):
    if start is None:
        return _merge_svec(gvlist, 0, len(gvlist))
    n = stop - start
    if n <= 0:
        return gvar.svec(0)
    if n == 1:
        return _getsvec(gvlist[start])
    left = _merge_svec(gvlist, start, start + n // 2)
    right = _merge_svec(gvlist, start + n // 2, stop)
    return left.add(right, 1, 1)

def jacobian(g):
    """
    Extract the jacobian of gvars w.r.t. primary gvars.
    
    Parameters
    ----------
    g : array_like
        An array of numbers or gvars.
    
    Returns
    -------
    jac : array
        The shape is g.shape + (m,), where m is the total number of primary
        gvars that g depends on.
    indices : (m,) int array
        The indices that map the last axis of jac to primary gvars in the
        global covariance matrix.

    See also
    --------
    from_jacobian
    """
    g = numpy.asarray(g)
    v = _merge_svec(g.flat)
    indices = v.indices()
    jac = numpy.zeros((g.size, len(indices)), float)
    for i, x in enumerate(g.flat):
        v = _getsvec(x)
        ind = numpy.searchsorted(indices, v.indices())
        jac[i, ind] = v.values()
    jac = jac.reshape(g.shape + indices.shape)
    return jac, indices

def from_jacobian(mean, jac, indices):
    """
    Create new gvars from a jacobian w.r.t. primary gvars.
    
    Parameters
    ----------
    mean : array_like
        An array of numbers with the means of the new gvars.
    jac : mean.shape + (m,) array
        The derivatives of each new gvar w.r.t. m primary gvars.
    indices : (m,) int array
        The indices of the primary gvars.
    
    Returns
    -------
    g : mean.shape array
        The new gvars.

    See also
    --------
    jacobian
    """
    cov = gvar.gvar.cov
    mean = numpy.asarray(mean)
    shape = mean.shape
    mean = mean.flat
    jac = numpy.asarray(jac)
    jac = jac.reshape(len(mean), len(indices))
    g = numpy.zeros(len(mean), object)
    for i, jacrow in enumerate(jac):
        der = gvar.svec(len(indices))
        der._assign(jacrow, indices)
        g[i] = gvar.GVar(mean[i], der, cov)
    return g.reshape(shape)



# lsqfitgp/_Kernel/_alg.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" register algops on CrossKernel and AffineSpan """

import functools

from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _special

from . import _util
from ._crosskernel import CrossKernel, AffineSpan

@CrossKernel.register_algop
def add(tcls, self, other):
    r"""
    
    Sum of kernels.
    
    .. math::
        \mathrm{newkernel}(x, y) &= \mathrm{kernel}(x, y) + \mathrm{other}(x, y), \\
        \mathrm{newkernel}(x, y) &= \mathrm{kernel}(x, y) + \mathrm{other}.
    
    Parameters
    ----------
    other : CrossKernel or scalar
        The other kernel.
    
    """
    core = self.core
    if _util.is_numerical_scalar(other):
        newcore = lambda x, y, **kw: core(x, y, **kw) + other
    elif isinstance(other, CrossKernel):
        other = other.core
        newcore = lambda x, y, **kw: core(x, y, **kw) + other(x, y, **kw)
    else:
        return NotImplemented
    return self._clone(core=newcore)

@CrossKernel.register_algop
def mul(tcls, self, other):
    r"""
    
    Product of kernels.
    
    .. math::
        \mathrm{newkernel}(x, y) &= \mathrm{kernel}(x, y) \cdot \mathrm{other}(x, y), \\
        \mathrm{newkernel}(x, y) &= \mathrm{kernel}(x, y) \cdot \mathrm{other}.
    
    Parameters
    ----------
    other : CrossKernel or scalar
        The other kernel.
    
    """
    core = self.core
    if _util.is_numerical_scalar(other):
        newcore = lambda x, y, **kw: core(x, y, **kw) * other
    elif isinstance(other, CrossKernel):
        other = other.core
        newcore = lambda x, y, **kw: core(x, y, **kw) * other(x, y, **kw)
    else:
        return NotImplemented
    return self._clone(core=newcore)

@CrossKernel.register_algop
def pow(tcls, self, *, exponent):
    r"""
    
    Power of the kernel.
    
    .. math::
        \mathrm{newkernel}(x, y) = \mathrm{kernel}(x, y)^{\mathrm{exponent}}
    
    Parameters
    ----------
    exponent : nonnegative integer
        The exponent. If traced by jax, it must have unsigned integer type.
    
    """
    if _util.is_nonnegative_integer_scalar(exponent):
        core = self.core
        newcore = lambda x, y, **kw: core(x, y, **kw) ** exponent
        return self._clone(core=newcore)
    else:
        return NotImplemented

    # TODO this will raise TypeError on negative integers. It should stop
    # method search and raise ValueError. Same for rpow. Check if it is a
    # scalar, then check if it satisfies the bound.

@CrossKernel.register_algop
def rpow(tcls, self, *, base):
    r"""
    
    Exponentiation of the kernel.
    
    .. math::
        \text{newkernel}(x, y) = \text{base}^{\text{kernel}(x, y)}
    
    Parameters
    ----------
    base : scalar
        A number >= 1. If traced by jax, the value is not checked.
    
    """
    if _util.is_scalar_cond_trueontracer(base, lambda x: x >= 1):
        core = self.core
        newcore = lambda x, y, **kw: base ** core(x, y, **kw)
        return self._clone(core=newcore)
    else:
        return NotImplemented

CrossKernel.register_ufuncalgop(jnp.tan)
# CrossKernel.register_ufuncalgop(lambda x: 1 / jnp.sinc(x), '1/sinc')
CrossKernel.register_ufuncalgop(lambda x: 1 / jnp.cos(x), '1/cos')
CrossKernel.register_ufuncalgop(jnp.arcsin)
CrossKernel.register_ufuncalgop(lambda x: 1 / jnp.arccos(x), '1/arccos')
CrossKernel.register_ufuncalgop(lambda x: 1 / (1 - x), '1/(1-x)')
CrossKernel.register_ufuncalgop(jnp.exp)
CrossKernel.register_ufuncalgop(lambda x: -jnp.log1p(-x), '-log1p(-x)')
CrossKernel.register_ufuncalgop(jnp.expm1)
CrossKernel.register_ufuncalgop(_special.expm1x)
CrossKernel.register_ufuncalgop(jnp.sinh)
CrossKernel.register_ufuncalgop(jnp.cosh)
CrossKernel.register_ufuncalgop(jnp.arctanh)
CrossKernel.register_ufuncalgop(jspecial.i0)
CrossKernel.register_ufuncalgop(jspecial.i1)
# @CrossKernel.register_ufuncalgop
# def iv(x, *, order):
#     assert _util.is_nonnegative_scalar_trueontracer(order)
#     return _special.iv(order, x)

# TODO other unary algop:
# - hypergeom (wrap the scipy impl in _special)

@functools.partial(AffineSpan.register_algop, transfname='add')
def affine_add(tcls, self, other):
    newself = AffineSpan.super_transf('add', self, other)
    if _util.is_numerical_scalar(other):
        dynkw = dict(self.dynkw)
        dynkw['offset'] = dynkw['offset'] + other
        return newself._clone(self.__class__, dynkw=dynkw)
    else:
        return newself

@functools.partial(AffineSpan.register_algop, transfname='mul')
def affine_mul(tcls, self, other):
    newself = AffineSpan.super_transf('mul', self, other)
    if _util.is_numerical_scalar(other):
        dynkw = dict(self.dynkw)
        dynkw['offset'] = other * dynkw['offset']
        dynkw['ampl'] = other * dynkw['ampl']
        return newself._clone(self.__class__, dynkw=dynkw)
    else:
        return newself



# lsqfitgp/_kernels/_bart.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import jax
from jax import numpy as jnp
from jax import lax
from jax.scipy import special as jspecial
from numpy.lib import recfunctions

from .. import _jaxext
from .. import _array
from .._Kernel import kernel

@kernel(derivable=False, batchbytes=10e6)
# TODO maybe batching should be done automatically by GP instead of by the
# kernels? But before doing that I need to support batching non-traceable
# functions.
def _BARTBase(x, y,
    alpha=0.95,
    beta=2,
    maxd=2,
    gamma=1,
    splits=None,
    pnt=None,
    intercept=True,
    weights=None,
    reset=None,
    indices=False):
    """
    BART kernel.

    Good default parameters: ``gamma=0.95``; ``maxd=4, reset=2`` if not fitting
    the hyperparameters (``alpha`` and ``beta``), ``maxd=10, reset=[2,4,6,8]``
    otherwise. Derivatives are faster with forward autodiff.
    
    Parameters
    ----------
    x, y : arrays
        Input points. The array type can be structured, in which case every leaf
        field represents a dimension; or unstructured, which specifies a single
        dimension.
    alpha, beta : scalar
        The parameters of the branching probability.
    maxd : int
        The maximum depth of the trees.
    splits : pair of arrays
        The first is an int (p,) array containing the number of splitting
        points along each dimension, the second has shape (n, p) and contains
        the sorted splitting points in each column, filled with high values
        after the length.
    gamma : scalar or str
        Interpolation coefficient in [0, 1] between a lower and a upper
        bound on the infinite maxd limit, or a string 'auto' indicating to
        use a formula which depends on alpha, beta, maxd and the number of
        covariates, empirically calibrated on maxd from 1 to 3. Default 1
        (upper bound).
    pnt : (maxd + 1,) array, optional
        Nontermination probabilities at depths 0...maxd. If specified,
        ``alpha``, ``beta`` and ``maxd`` are ignored.
    intercept : bool
        The correlation is in [1 - alpha, 1] (or [1 - pnt[0], 1] when using
        pnt). If intercept=False, it is rescaled to [0, 1]. Default True.
    weights : (p,) array
        Unnormalized selection probabilities for the covariate axes. If not
        specified, all axes have the same probability to be selected for
        splitting.
    reset : int or sequence of int, optional
        List of depths at which the recursion is reset, in the sense that the
        function value at a reset depth is evaluated on the initial inputs for
        all recursion paths, instead of the modified input handed down by the
        recursion. Default none.
    indices : bool
        If False (default), the inputs ``x``, ``y`` represent coordinate values.
        If True, they are taken to be already the indices of the points in the
        splitting grid, as can be obtained with `BART.indices_from_coord`.
    
    Methods
    -------
    splits_from_coord
    indices_from_coord
    correlation
    
    Notes
    -----
    This is the covariance function of the latent mean prior of BART (Bayesian
    Additive Regression Trees) [1]_ in the limit of an infinite number of
    trees, and with an upper bound :math:`D` on the depth of the trees. This
    prior is the distribution of the function
    
    .. math::
        f(\\mathbf x) = \\lim_{m\\to\\infty}
        \\sum_{j=1}^m g(\\mathbf x; T_j, M_j),
    
    where each :math:`g(\\mathbf x; T_j, M_j)` is a decision tree evaluated at
    :math:`\\mathbf x`, with structure :math:`T_j` and leaf values :math:`M_j`.
    The trees are i.i.d., with the following distribution for :math:`T_j`: for
    a node at depth :math:`d`, with :math:`d = 0` for the root, the probability
    of not being a leaf, conditional on its existence and its ancestors only, is
    
    .. math::
        P_d = \\alpha (1+d)^{-\\beta}, \\quad
        \\alpha \\in [0, 1], \\quad \\beta \\ge 0.
    
    For a non-leaf node, conditional on existence and ancestors, the splitting
    variable has uniform distribution amongst the variables with any splitting
    points not used by ancestors, and the splitting point has uniform
    distribution amongst the available ones. The splitting points are fixed,
    tipically from the data.
    
    The distribution of leaves :math:`M_j` is i.i.d. Normal with variance
    :math:`1/m`, such that :math:`f(x)` has variance 1. In the limit
    :math:`m\\to\\infty`, the distribution of :math:`f(x)` becomes a Gaussian
    process.
    
    Since the trees are independent, the covariance function can be computed
    for a single tree. Consider two coordinates :math:`x` and :math:`y`, with
    :math:`x \\le y`. Let :math:`n^-`, :math:`n^0` and :math:`n^+` be the
    number of splitting points respectively before :math:`x`, between
    :math:`x`, :math:`y` and after :math:`y`. Next, define :math:`\\mathbf
    n^-`, :math:`\\mathbf n^0` and :math:`\\mathbf n^+` as the vectors of such
    quantities for each dimension, with a total of :math:`p` dimensions, and
    :math:`\\mathbf n = \\mathbf n^- + \\mathbf n^0 + \\mathbf n^+`. Then the
    covariance function can be written recursively as
    
    .. math::
        \\newcommand{\\nvecs}{\\mathbf n^-, \\mathbf n^0, \\mathbf n^+}
        k(\\mathbf x, \\mathbf y) &= k_0(\\nvecs), \\\\
        k_D(\\nvecs) &= 1 - (1 - \\gamma) P_D,
            \\quad \\mathbf n^0 \\ne \\mathbf 0, \\\\
        k_d(\\mathbf 0, \\mathbf 0, \\mathbf 0) &= 1, \\\\
        k_d(\\nvecs) &= 1 - P_d \\Bigg(1 - \\frac1{W(\\mathbf n)}
            \\sum_{\\substack{i=1 \\\\ n_i\\ne 0}}^p
                \\frac{w_i}{n_i} \\Bigg( \\\\
                &\\qquad \\sum_{k=0}^{n^-_i - 1}
                k_{d+1}(\\mathbf n^-_{n^-_i=k}, \\mathbf n^0, \\mathbf n^+)
                + {} \\\\
                &\\qquad \\sum_{k=0}^{n^+_i - 1}
                k_{d+1}(\\mathbf n^-, \\mathbf n^0, \\mathbf n^+_{n^+_i=k})
            \\Bigg)
        \\Bigg), \\quad d < D, \\\\
        W(\\mathbf n) &= \\sum_{\\substack{i=1 \\\\ n_i\\ne 0}}^p w_i.
        
    The introduction of a maximum depth :math:`D` is necessary for
    computational feasibility. As :math:`D` increases, the result converges to
    the one without depth limit. For :math:`D \\le 2` (the default value), the
    covariance is implemented in closed form and takes :math:`O(p)` to compute.
    For :math:`D > 2`, the computational complexity grows exponentially as
    :math:`O(p(\\bar np)^{D-2})`, where :math:`\\bar n` is the average number of splitting
    points along a dimension.
    
    In the maximum allowed depth is 1, i.e., either :math:`D = 1` or
    :math:`\\beta\\to\\infty`, the kernel assumes the simple form
    
    .. math::
        k(\\mathbf x, \\mathbf y) &= 1 - P_0 \\left(
            1 - Q + \\frac Q{W(\\mathbf n)}
            \\sum_{\\substack{i=1 \\\\ n_i\\ne 0}}^p w_i
            \\frac{n^0_i}{n_i} \\right), \\\\
        Q &= \\begin{cases}
            1 - (1 - \\gamma) P_1 & \\mathbf n^0 \\ne \\mathbf 0, \\\\
            1 & \\mathbf n^0 = \\mathbf 0,
        \\end{cases}
    
    which is separable along dimensions, i.e., it has no interactions.
    
    References
    ----------
    .. [1] Hugh A. Chipman, Edward I. George, Robert E. McCulloch "BART:
        Bayesian additive regression trees," The Annals of Applied Statistics,
        Ann. Appl. Stat. 4(1), 266-298, (March 2010).
    """

    splits = BART._check_splits(splits, indices)
    if not x.dtype.names:
        x = x[..., None]
    if not y.dtype.names:
        y = y[..., None]
    if indices:
        ix = BART._check_x(x)
        iy = BART._check_x(y)
    else:
        ix = BART._indices_from_coord(x, splits)
        iy = BART._indices_from_coord(y, splits)
    return BART.correlation(
        splits[0], ix, iy,
        pnt=pnt, alpha=alpha, beta=beta, gamma=gamma, maxd=maxd,
        intercept=intercept, weights=weights, reset=reset, altinput=True,
    )
    
    # TODO
    # - approximate as stationary w.r.t. indices (is it pos def?)
    # - default gamma='auto'? => wait for better gamma
    # - make gamma='auto' depend on maxd and reset with a dictionary, error
    #   if not specified
    # - do not require to specify splitting points if using indices

class BART(_BARTBase):
    
    __doc__ = _BARTBase.__doc__
    
    @classmethod
    def splits_from_coord(cls, x):
        """
        Generate splitting points from data.
        
        Parameters
        ----------
        x : array of numbers
            The data. Can be passed in two formats: 1) a structured array where
            each leaf field represents a dimension, 2) a normal array where the
            last axis runs over dimensions. In the structured case, each
            index in any shaped field is a different dimension.
        
        Returns
        -------
        length : int (p,) array
            The number of splitting points along each of ``p`` dimensions.
        splits : (n, p) array
            Each column contains the sorted splitting points along a dimension.
            The splitting points are the midpoints between consecutive values
            appearing in `x` for that dimension. Column ``splits[:, i]``
            contains splitting points only up to ``length[i]``, while afterward
            it is filled with a very large value.
        
        """
        x = cls._check_x(x)
        return cls._splits_from_coord(x)
        
        # TODO options like BayesTree, i.e., use an evenly spaced range
        # instead of quantilizing, and set a maximum number of splits. Use the
        # same parameter names as BayesTree::bart, but change the defaults.

    @staticmethod
    @jax.jit
    def _splits_from_coord(x):
        """
        Jitted implementation of splits_from_coord. Applying jit avoids the
        recompilation in lax.scan each time the method is called, and
        splits_from_coord can not be jitted directly because x could be a numpy
        structured array.
        """
        x = x.reshape(-1, x.shape[-1]) if x.size else x.reshape(1, x.shape[-1])
        if jnp.issubdtype(x.dtype, jnp.inexact):
            info = jnp.finfo
        else:
            info = jnp.iinfo
        fill = info(x.dtype).max
        def loop(_, xi):
            u = jnp.unique(xi, size=xi.size, fill_value=fill)
            m = jnp.where(u[1:] < fill, (u[1:] + u[:-1]) / 2, fill)
            l = jnp.searchsorted(m, fill)
            return _, (l, m)
        _, (length, midpoints) = lax.scan(loop, None, x.T)
        return length, midpoints.T
    
    @classmethod
    def indices_from_coord(cls, x, splits):
        """
        Convert coordinates to indices w.r.t. splitting points.
        
        Parameters
        ----------
        x : array of numbers
            The coordinates. Can be passed in two formats: 1) a structured
            array where each leaf field represents a dimension, 2) a normal
            array where the last axis runs over dimensions. In the structured
            case, each index in any shaped field is a different dimension.
        splits : pair of arrays
            The first is an int (p,) array containing the number of splitting
            points along each dimension, the second has shape (n, p) and
            contains the sorted splitting points in each column, filled with
            high values after the length.
        
        Returns
        -------
        ix : int array
            An array with the same shape as ``x``, unless ``x`` is a structured
            array, in which case the last axis of ``ix`` is the flattened version
            of the structured type. ``ix`` contains indices mapping ``x`` to
            positions between splitting points along each coordinate, with the
            following convention: index 0 means before the first split, index
            i > 0 means between split i - 1 and split i.
        
        """
        splits = cls._check_splits(splits, False)
        return cls._indices_from_coord(x, splits)

    @classmethod
    def _indices_from_coord(cls, x, checked_splits):
        x = cls._check_x(x)
        if x.shape[-1] != checked_splits[0].size:
            raise ValueError(f'splitting grid is for {checked_splits[0].size} '
                f'dimensions, found {x.shape[-1]}')
        return cls._searchsorted_vectorized(checked_splits[1], x)
    
    @classmethod
    def correlation(cls,
        splitsbefore_or_totalsplits,
        splitsbetween_or_index1,
        splitsafter_or_index2,
        *,
        alpha=0.95,
        beta=2,
        gamma=1,
        maxd=2,
        debug=False,
        pnt=None,
        intercept=True,
        weights=None,
        reset=None,
        altinput=False):
        """
        Compute the BART prior correlation between two points.

        Apart from arguments ``maxd``, ``debug`` and ``reset``, this method is fully
        vectorized.
    
        Parameters
        ----------
        splitsbefore_or_totalsplits : int (p,) array
            The number of splitting points less than the two points, separately
            along each coordinate, or the total number of splits if ``altinput``.
        splitsbetween_or_index1 : int (p,) array
            The number of splitting points between the two points, separately
            along each coordinate, or the index in the splitting bins of the
            first point if ``altinput``, where 0 means to the left of the leftmost
            splitting point.
        splitsafter_or_index2 : int (p,) array
            The number of splitting points greater than the two points,
            separately along each coordinate, or the index in the splitting bins
            of the second point if ``altinput``.
        debug : bool
            If True, disable shortcuts in the tree recursion. Default False.
        altinput : bool
            If True, take as input the indices in the splitting bins of the
            points instead of the counts of splitting points separating them,
            and use a different implementation optimized for that case. Default
            False. The `BART` kernel uses ``altinput=True``.
        Other parameters :
            See `BART`.

        Returns
        -------
        corr : scalar
            The prior correlation.
        """
        
        # check splitting indices are integers
        splitsbefore_or_totalsplits = jnp.asarray(splitsbefore_or_totalsplits)
        splitsbetween_or_index1 = jnp.asarray(splitsbetween_or_index1)
        splitsafter_or_index2 = jnp.asarray(splitsafter_or_index2)
        assert jnp.issubdtype(splitsbefore_or_totalsplits.dtype, jnp.integer)
        assert jnp.issubdtype(splitsbetween_or_index1.dtype, jnp.integer)
        assert jnp.issubdtype(splitsafter_or_index2.dtype, jnp.integer)
        
        # check splitting indices
        with _jaxext.skipifabstract():
            assert jnp.all(splitsbefore_or_totalsplits >= 0), 'splitting counts must be nonnegative'
            if altinput:
                assert jnp.all((0 <= splitsbetween_or_index1) & (splitsbetween_or_index1 <= splitsbefore_or_totalsplits)), 'splitting index must be in [0, n]'
                assert jnp.all((0 <= splitsafter_or_index2) & (splitsafter_or_index2 <= splitsbefore_or_totalsplits)), 'splitting index must be in [0, n]'
            else:
                assert jnp.all(splitsbetween_or_index1 >= 0), 'splitting counts must be nonnegative'
                assert jnp.all(splitsafter_or_index2 >= 0), 'splitting counts must be nonnegative'
        
        # get splitting probabilities
        if pnt is None:
            assert maxd == int(maxd) and maxd >= 0, maxd
            alpha = jnp.asarray(alpha)
            beta = jnp.asarray(beta)
            with _jaxext.skipifabstract():
                assert jnp.all((0 <= alpha) & (alpha <= 1)), 'alpha must be in [0, 1]'
                assert jnp.all(beta >= 0), 'beta must be in [0, inf)'
            d = jnp.arange(maxd + 1)
            alpha = alpha[..., None]
            beta = beta[..., None]
            pnt = alpha / (1 + d) ** beta
        else:
            pnt = jnp.asarray(pnt)
        
        # get covariate weights
        if weights is None:
            weights = jnp.ones(splitsbefore_or_totalsplits.shape[-1], pnt.dtype)
        else:
            weights = jnp.asarray(weights)
        
        # get interpolation coefficients
        if isinstance(gamma, str):
            if gamma == 'auto':
                assert reset is None and 1 <= pnt.shape[-1] - 1 <= 3
                p = weights.shape[-1]
                gamma = cls._gamma(p, pnt)
            else:
                raise KeyError(gamma)
        else:
            gamma = jnp.asarray(gamma)
        
        # check values are in range
        with _jaxext.skipifabstract():
            assert jnp.all((0 <= gamma) & (gamma <= 1)), 'gamma must be in [0, 1]'
            assert jnp.all((0 <= pnt) & (pnt <= 1)), 'pnt must be in [0, 1]'
            assert jnp.all(weights >= 0), 'weights must be in [0, inf)'

        # set first splitting probability to 1 to remove flat baseline (keep
        # last!)
        if not intercept:
            pnt = pnt.at[..., 0].set(1)
        
        # expand and check recursion reset depths
        if reset is None:
            reset = []
        if not hasattr(reset, '__len__'):
            reset = [reset]
        reset = [0] + list(reset) + [pnt.shape[-1] - 1]
        for i, j in zip(reset, reset[1:]):
            assert int(j) == j and i <= j, (i, j)

        # convert reset depths list to brackets with repetition
        brackets_norep = list(zip(reset, reset[1:]))
        brackets = [brackets_norep[0] + (1,)]
        for t, b in brackets_norep[1:]:
            lt, lb, lr = brackets[-1]
            if altinput and not debug and lr * (b - t) == lb - lt and b - t <= 2:
                brackets[-1] = lt, b, lr + 1
            else:
                brackets.append((t, b, 1))

        # call recursive function for each recursion slice
        corr = gamma
        for t, b, repeat in reversed(brackets):
            probs = pnt[..., t:b + 1]
            if t > 0:
                probs = probs.at[..., 0].set(1)
            if repeat > 1:
                head = probs[..., 0:1]
                one = jnp.ones_like(head)
                probs = jnp.concatenate(sum(reversed([
                    [head if i == 0 else one, p]
                    for i, p in enumerate(jnp.split(probs[..., 1:], repeat, axis=-1))
                ]), start=[]), axis=-1)
            else:
                repeat = None
            corr = cls._correlation_vectorized(
                splitsbefore_or_totalsplits,
                splitsbetween_or_index1,
                splitsafter_or_index2,
                probs, corr, weights,
                debug, altinput, repeat,
            )
        return corr
    
    # TODO public method to compute pnt
        
    @staticmethod
    def _gamma(p, pnt):
        # gamma(alpha, beta, maxd) =
        #   = (gamma_0 - gamma_d maxd) (1 - alpha^s 2^(-t beta)) =
        #   = (gamma_0 - gamma_d maxd) (1 - P0^s-t P1^t)

        gamma_0 = 0.611 + 0.021 * jnp.exp(-1.3 * (p - 1))
        gamma_d = -0.0034 + 0.084 * jnp.exp(-2.02 * (p - 1))
        s = 2.03 - 0.69 * jnp.exp(-0.72 * (p - 1))
        t = 4.01 - 1.49 * jnp.exp(-0.77 * (p - 1))

        maxd = pnt.shape[-1] - 1
        floor = jnp.clip(gamma_0 - gamma_d * maxd, 0, 1)

        P0 = pnt[..., 0]
        P1 = jnp.minimum(P0, pnt[..., 1])
        corner = jnp.where(P0, 1 - P0 ** (s - t) * P1 ** t, 1)
        
        return floor * corner
        
        # TODO make this public?

    @staticmethod
    def _check_x(x):
        x = _array.asarray(x)
        if x.dtype.names:
            x = recfunctions.structured_to_unstructured(x)
        return x

    @staticmethod
    def _check_splits(splits, indices):
        l, s = splits
        l = jnp.asarray(l)
        assert l.ndim == 1
        if not indices:
            s = jnp.asarray(s)
            assert 1 <= s.ndim <= 2
            if s.ndim == 1:
                s = s[:, None]
            assert l.size == s.shape[1]
        with _jaxext.skipifabstract():
            assert jnp.all((0 <= l) & (l <= s.shape[0])), 'length out of bounds'
            if not indices:
                assert jnp.all(jnp.sort(s, axis=0) == s), 'unsorted splitting points'
        return l, s
    
    @staticmethod
    @functools.partial(jax.jit, static_argnames=('side',))
    def _searchsorted_vectorized(A, V, **kw):
        """
        A : (n, p)
        V : (..., p)
        out : (..., p)
        """
        def loop(_, av):
            return _, jnp.searchsorted(*av, **kw)
        _, out = lax.scan(loop, None, (A.T, V.T))
        return out.T

    @classmethod
    @functools.partial(jax.jit, static_argnums=(0, 7))
    def _correlation_old(cls, nminus, n0, nplus, pnt, gamma, w, debug):
        """ old version, kept around for cross-checking """

        assert nminus.shape == n0.shape == nplus.shape == w.shape
        assert nminus.ndim == 1 and nminus.size >= 0
        assert pnt.ndim == 1 and pnt.size > 0
        # TODO repeat this shape checks in BART.correlation such that the
        # error messages are user-legible

        # optimization to avoid looping over ignored axes
        nminus = jnp.where(w, nminus, 0)
        n0 = jnp.where(w, n0, 0)
        nplus = jnp.where(w, nplus, 0)
        
        float_type = _jaxext.float_type(pnt, gamma, w)
        
        if nminus.size == 0:
            return jnp.array(1, float_type)
        
        anyn0 = jnp.any(jnp.logical_and(n0, w))

        if pnt.size == 1:
            return jnp.where(anyn0, 1 - (1 - gamma) * pnt[0], 1)
    
        nout = nminus + nplus
        n = nout + n0
        Wn = jnp.sum(jnp.where(n, w, 0)) # <-- @

        if pnt.size == 2 and not debug:
            Q = 1 - (1 - gamma) * pnt[1]
            sump = Q * jnp.sum(jnp.where(n, w * nout / n, 0)) # <-- @
            return jnp.where(anyn0, 1 - pnt[0] * (1 - sump / Wn), 1)
    
        if pnt.size == 3 and not debug:
            Q = 1 - (1 - gamma) * pnt[2]
            s = w * nout / n
            S = jnp.sum(jnp.where(n, s, 0)) # <-- @
            t = w * n0 / n
            psin = jspecial.digamma(n.astype(float_type))
            def terms(nminus, nplus):
                nminus0 = nminus + n0
                Wnmod = Wn - jnp.where(nminus0, 0, w)
                frac = jnp.where(nminus0, w * nminus / nminus0, 0)
                terms1 = (S - s + frac) / Wnmod
                psi1nminus0 = jspecial.digamma((1 + nminus0).astype(float_type))
                terms2 = ((nplus - 1) * (S + t) - w * n0 * (psin - psi1nminus0)) / Wn
                return jnp.where(nplus, terms1 + terms2, 0)
            tplus = terms(nminus, nplus)
            tminus = terms(nplus, nminus)
            tall = jnp.where(n, w * (tplus + tminus) / n, 0)
            sump = (1 - pnt[1]) * S + pnt[1] * Q * jnp.sum(tall) # <-- @
            return jnp.where(anyn0, 1 - pnt[0] * (1 - sump / Wn), 1)
        
            # TODO the pnt.size == 3 calculation is probably less accurate than
            # the recursive one, see comparison limits > 30 ULP in test_bart.py
    
        p = len(nminus)

        val = (0., nminus, n0, nplus)
        def loop(i, val):
            sump, nminus, n0, nplus = val

            nminusi = nminus[i]
            n0i = n0[i]
            nplusi = nplus[i]
            ni = nminusi + n0i + nplusi
        
            val = (0., nminus, n0, nplus, i, nminusi)
            def loop(k, val):
                sumn, nminus, n0, nplus, i, nminusi = val
            
                # here I use the fact that .at[].set won't set the value if the
                # index is out of bounds
                nminus = nminus.at[jnp.where(k < nminusi, i, i + p)].set(k)
                nplus = nplus.at[jnp.where(k >= nminusi, i, i + p)].set(k - nminusi)
            
                sumn += cls._correlation_old(nminus, n0, nplus, pnt[1:], gamma, w, debug)
            
                nminus = nminus.at[i].set(nminusi)
                nplus = nplus.at[i].set(nplusi)
            
                return sumn, nminus, n0, nplus, i, nminusi
        
            # if ni == 0 I skip recursion by passing 0 as iteration end
            end = jnp.where(ni, nminusi + nplusi, 0)
            start = jnp.zeros_like(end)
            sumn, nminus, n0, nplus, _, _ = lax.fori_loop(start, end, loop, val)

            sump += jnp.where(ni, w[i] * sumn / ni, 0)

            return sump, nminus, n0, nplus

        # skip summation if all(n0 == 0)
        end = jnp.where(anyn0, p, 0)
        sump, _, _, _ = lax.fori_loop(0, end, loop, val)

        return jnp.where(anyn0, 1 - pnt[0] * (1 - sump / Wn), 1)

    @staticmethod
    def _scan_but_first(f, init, xs):
        """ lax.scan, but execute separately the first cycle. The point is that
        I use it when the first cycle works on smaller arrays due to
        broadcasting. """
        assert isinstance(xs, jnp.ndarray)
        assert len(xs) > 0
        init, out = f(init, xs[0])
        assert out is None
        if len(xs) == 1:
            return init, out
        elif len(xs) == 2:
            return f(init, xs[1])
        else:
            return lax.scan(f, init, xs[1:])

    @classmethod
    @functools.partial(jax.jit, static_argnums=(0, 7, 8))
    def _correlation(cls, n, ix, iy, pnt, gamma, w, debug, repeat):
        # this implementation is optimized assuming that the shapes are as
        # follows:
        #   n     (p,)
        #   ix    (n, 1, p)
        #   iy    (1, n, p)
        #   pnt   (d,)
        #   gamma () or (n, n)
        #   w     (p,)

        assert n.ndim == 1
        assert n.shape == ix.shape == iy.shape == w.shape
        assert pnt.ndim == 1 and pnt.size > 0
        assert gamma.ndim == 0
        # TODO repeat this shape checks in BART.correlation such that the
        # error messages are user-legible

        # check the strict conditions under which `repeat` is implemented
        if repeat is not None:
            assert (
                not debug
                and repeat > 0
                and pnt.size % repeat == 0
                and pnt.size // repeat <= 3
            )
        else:
            repeat = 1

        # infer float type from float arguments
        flt = _jaxext.float_type(pnt, gamma, w)
        
        # no covariates, always return 1
        if n.size == 0:
            return jnp.array(1, flt)
        
        # pre-cast all floats to the common type, to avoid unwanted float32
        # calculations in mixed float-integer operations
        pnt = pnt.astype(flt)
        gamma = gamma.astype(flt)
        w = w.astype(flt)
        
        # ignore zero-weight axes
        n = jnp.where(w, n, 0)
        ix = jnp.where(w, ix, 0)
        iy = jnp.where(w, iy, 0)

        # check if the points coincide
        seed = jnp.uint64(16132933535611723338)
        hx = _jaxext.fasthash64(ix, seed)
        hy = _jaxext.fasthash64(iy, seed)
        anyn0 = hx != hy
        # no hash collision checking, it would be branchless because of vmap,
        # the probability of collision building a nxn matrix with n=10000 is
        # -expm1(10000**2 * log1p(-1/2**64)) = 5e-12.

        # base case of the recursion, no dependence on points apart from the
        # case when they are equal
        if pnt.size // repeat == 1:
            def loop(carry, pnt):
                anyn0, gamma = carry
                gamma = jnp.where(anyn0, 1 - (1 - gamma) * pnt[0], 1)
                return (anyn0, gamma), None
            (_, gamma), _ = cls._scan_but_first(loop, (anyn0, gamma), pnt.reshape(repeat, -1))
            return gamma
        
        # normalization for axes weights
        Wn = jnp.sum(jnp.where(n, w, 0))

        # shortcut for the last two levels of the recursion
        if pnt.size // repeat == 2 and not debug:
            n0 = jnp.abs(ix - iy)
            sum_term = jnp.where(n, w / n, 0) @ n0
            def loop(carry, pnt):
                anyn0, Wn, sum_term, gamma = carry
                Q = 1 - pnt[1] + gamma * pnt[1]
                P0 = pnt[0]
                result = 1 - P0 + Q * (P0 - P0 / Wn * sum_term)
                gamma = jnp.where(anyn0, result, 1)
                return (anyn0, Wn, sum_term, gamma), None
            (_, _, _, gamma), _ = cls._scan_but_first(loop, (anyn0, Wn, sum_term, gamma), pnt.reshape(repeat, -1))
            return gamma
    
        # convert to alternative format
        xlty = ix < iy
        minxy = jnp.where(xlty, ix, iy)
        maxxy = jnp.where(xlty, iy, ix)
        n0 = maxxy - minxy
        
        # shortcut for the last three levels of the recursion
        if pnt.size // repeat == 3 and not debug:
            nminus0 = maxxy
            nplus0 = n - minxy
            nout = n - n0

            inv_Wn = 1 / Wn
            inv_Wnmod = 1 / (Wn - jnp.where(n, w, 0))
            inv_Wnminus = jnp.where(nplus0, inv_Wn, inv_Wnmod)
            inv_Wnplus = jnp.where(nminus0, inv_Wn, inv_Wnmod)
            wn = jnp.where(n, w / n, 0)
            S = wn @ nout

            t = wn * n0
            terms1 = (S + t) * (inv_Wnminus + inv_Wnplus + inv_Wn * (nout - 2))

            terms2  = jnp.where( nplus0, w * inv_Wn * n0 /  nplus0, w * inv_Wnmod)
            terms2 += jnp.where(nminus0, w * inv_Wn * n0 / nminus0, w * inv_Wnmod)

            psin = jspecial.digamma(jnp.where(n, n, 1).astype(flt))
            psiminus = jnp.where(xlty,
                jspecial.digamma((1 + iy).astype(flt)),
                jspecial.digamma((1 + ix).astype(flt)),
            )
            psiplus = jnp.where(xlty,
                jspecial.digamma((1 + n - ix).astype(flt)),
                jspecial.digamma((1 + n - iy).astype(flt)),
            )
            terms3 = w * inv_Wn * n0 * (2 * psin - psiminus - psiplus)

            terms = terms1 - terms2 - terms3
            sumi = wn @ terms

            def loop(carry, pnt):
                anyn0, inv_Wn, S, sumi, gamma = carry
                Q = 1 + pnt[2] * (gamma - 1)
                sump = S + pnt[1] * (Q * sumi - S)
                result = 1 + pnt[0] * (inv_Wn * sump - 1)
                gamma = jnp.where(anyn0, result, 1)
                return (anyn0, inv_Wn, S, sumi, gamma), None
            (_, _, _, _, gamma), _ = cls._scan_but_first(loop, (anyn0, inv_Wn, S, sumi, gamma), pnt.reshape(repeat, -1))
            return gamma

        # finish conversion to alternative format
        nminus = minxy
        nplus = n - maxxy
        p = len(nminus)
        del ix, iy, maxxy, minxy

        val = (0., nminus, n0, nplus)
        def loop(i, val):
            sump, nminus, n0, nplus = val

            nminusi = nminus[i]
            n0i = n0[i]
            nplusi = nplus[i]
            ni = nminusi + n0i + nplusi
        
            val = (0., nminus, n0, nplus, i, nminusi)
            def loop(k, val):
                sumn, nminus, n0, nplus, i, nminusi = val
            
                # here I use the fact that .at[].set won't set the value if the
                # index is out of bounds
                nminus = nminus.at[jnp.where(k < nminusi, i, i + p)].set(k)
                nplus = nplus.at[jnp.where(k >= nminusi, i, i + p)].set(k - nminusi)
            
                n = nminus + n0 + nplus
                ix = nminus
                iy = nminus + n0
                sumn += cls._correlation(n, ix, iy, pnt[1:], gamma, w, debug, None)
            
                nminus = nminus.at[i].set(nminusi)
                nplus = nplus.at[i].set(nplusi)
            
                return sumn, nminus, n0, nplus, i, nminusi
        
            # if ni == 0 I skip recursion by passing 0 as iteration end
            end = jnp.where(ni, nminusi + nplusi, 0)
            start = jnp.zeros_like(end)
            sumn, nminus, n0, nplus, _, _ = lax.fori_loop(start, end, loop, val)

            sump += jnp.where(ni, w[i] * sumn / ni, 0)

            return sump, nminus, n0, nplus

        # skip summation if all(n0 == 0)
        end = jnp.where(anyn0, p, 0)
        sump, _, _, _ = lax.fori_loop(0, end, loop, val)

        return jnp.where(anyn0, 1 - pnt[0] * (1 - sump / Wn), 1)

    @classmethod
    @functools.partial(jnp.vectorize, excluded=(0, 7, 8, 9), signature='(p),(p),(p),(d),(),(p)->()')
    def _correlation_vectorized(cls, nminus_or_n, n0_or_ix, nplus_or_iy, pnt, gamma, w, debug, altinput, repeat):
        if altinput:
            func = lambda *args: cls._correlation(*args, repeat)
        else:
            func = cls._correlation_old
        return func(nminus_or_n, n0_or_ix, nplus_or_iy, pnt, gamma, w, bool(debug))



# lsqfitgp/_GP/_compute.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import warnings
import math

from jax import numpy as jnp
import numpy
import gvar

from .. import _linalg
from .. import _jaxext

from . import _base

class GPCompute(_base.GPBase):

    def __init__(self, *, solver, solverkw):
        self._decompcache = {} # tuple of keys -> Decomposition
        decomp = self._getdecomp(solver)
        self._decompclass = lambda K, **kwargs: decomp(K, **kwargs, **solverkw)

    def _clone(self):
        newself = super()._clone()
        newself._decompcache = self._decompcache.copy()
        newself._decompclass = self._decompclass
        return newself

    def _solver(self, keys, ycov=None, *, covtransf=None, **kw):
        """
        Return a decomposition of the covariance matrix of the keys in ``keys``
        plus the matrix ycov. Keyword arguments are passed to the decomposition.
        """
        
        # TODO cache ignores **kw.

        keys = tuple(keys)
        
        # Check if decomposition is in cache.
        if ycov is None:
            cache = self._decompcache.get(keys)
            if cache is not None:
                return cache
            # TODO use frozenset(keys) instead of tuple(keys) to make cache
            # work when order changes, but I have to permute the decomposition
            # to make that work. Needs an ad-hoc class in _linalg. Make the
            # decompcache a dict subclass that accepts tuples of keys but uses
            # internally frozenset.
        
        # Compute decomposition. # woodbury, currently un-implemented
        # if isinstance(ycov, _linalg.Decomposition):
        #     ancestors = []
        #     transfs = []
        #     for key in keys:
        #         elem = self._elements[key]
        #         nest = False
        #         if isinstance(elem, self._LinTransf):
        #             size = sum(self._elements[k].size for k in elem.keys)
        #             if size < elem.size:
        #                 nest = True
        #                 ancestors += list(elem.keys)
        #                 transfs.append(jnp.concatenate(elem.matrices(self), 1))
        #         if not nest:
        #             ancestors.append(key)
        #             transfs.append(jnp.eye(elem.size))
        #     transf = jlinalg.block_diag(*transfs)
        #     cov = self._assemblecovblocks(ancestors)
        #     if covtransf:
        #         ycov, transf, cov = covtransf((ycov, transf, cov))
        #     covdec = self._decompclass(cov, **kw)
        #     # TODO obtain covdec from _solver recursively, to use cache?
        #     decomp = _linalg.Woodbury2(ycov, transf, covdec, self._decompclass, sign=1, **kw)
        # else:
        Kxx = self._assemblecovblocks(keys)
        if ycov is not None:
            Kxx = Kxx + ycov
        if covtransf:
            Kxx = covtransf(Kxx)
        decomp = self._decompclass(Kxx, **kw)
        
        # Cache decomposition.
        if ycov is None:
            self._decompcache[keys] = decomp
        
        return decomp
        
    def _flatgiven(self, given, givencov):
        
        if not hasattr(given, 'keys'):
            raise TypeError('`given` must be dict')
        gcblack = givencov is None or isinstance(givencov, _linalg.Decomposition)
        if not gcblack and not hasattr(givencov, 'keys'):
            raise TypeError('`givenconv` must be None, dict or Decomposition')
        
        ylist = []
        keylist = []
        for key, l in given.items():
            if key not in self._elements:
                raise KeyError(key)
            
            if not isinstance(l, jnp.ndarray):
                # use numpy since there could be gvars
                l = numpy.asarray(l)
            shape = self._elements[key].shape
            if l.shape != shape:
                msg = 'given[{!r}] has shape {!r} different from shape {!r}'
                raise ValueError(msg.format(key, l.shape, shape))
            if l.dtype != object and not jnp.issubdtype(l.dtype, jnp.number):
                msg = 'given[{!r}] has non-numerical dtype {!r}'
                raise TypeError(msg.format(key, l.dtype))
            
            ylist.append(l.reshape(-1))
            keylist.append(key)
        
        # TODO error checking on the unpacking of givencov
        
        if gcblack:
            covblocks = givencov
        else:
            covblocks = [
                [
                    jnp.asarray(givencov[keylist[i], keylist[j]]).reshape(ylist[i].shape + ylist[j].shape)
                    for j in range(len(keylist))
                ]
                for i in range(len(keylist))
            ]
            
        return ylist, keylist, covblocks
    
    def pred(self, given, key=None, givencov=None, *, fromdata=None, raw=False, keepcorr=None):
        """
        
        Compute the posterior.
        
        The posterior can be computed either for all points or for a subset,
        and either directly from data or from a posterior obtained with a fit.
        The latter case is for when the Gaussian process was used in a fit with
        other parameters.
        
        The output is a collection of gvars, either an array or a dictionary
        of arrays. They are properly correlated with gvars returned by
        `prior` and with the input data/fit.
        
        The input is a dictionary of arrays, ``given``, with keys corresponding
        to the keys in the GP as added by `addx` or `addtransf`.
        
        Parameters
        ----------
        given : dictionary of arrays
            The data or fit result for some/all of the points in the GP.
            The arrays can contain either gvars or normal numbers, the latter
            being equivalent to zero-uncertainty gvars.
        key : None, key or list of keys, optional
            If None, compute the posterior for all points in the GP (also those
            used in ``given``). Otherwise only those specified by key.
        givencov : dictionary of arrays, optional
            Covariance matrix of ``given``. If not specified, the covariance
            is extracted from ``given`` with ``gvar.evalcov(given)``.
        fromdata : bool
            Mandatory. Specify if the contents of ``given`` are data or already
            a posterior.
        raw : bool, optional
            If True, instead of returning a collection of gvars, return
            the mean and the covariance. When the mean is a dictionary, the
            covariance is a dictionary whose keys are pairs of keys of the
            mean (the same format used by `gvar.evalcov`). Default False.
        keepcorr : bool, optional
            If True (default), the returned gvars are correlated with the
            prior and the data/fit. If False, they have the correct covariance
            between themselves, but are independent from all other preexisting
            gvars.
        
        Returns
        -------
        If raw=False (default):
        
        posterior : array or dictionary of arrays
            A collections of gvars representing the posterior.
        
        If raw=True:
        
        pmean : array or dictionary of arrays
            The mean of the posterior. Equivalent to ``gvar.mean(posterior)``.
        pcov : 2D array or dictionary of 2D arrays
            The covariance matrix of the posterior. If ``pmean`` is a
            dictionary, the keys of ``pcov`` are pairs of keys of ``pmean``.
            Equivalent to ``gvar.evalcov(posterior)``.
        
        """

        # TODO GP.pred(..., raw=True, onlyvariance=True) computes only the
        # variance (requires actually implementing diagquad at least in Chol and
        # Diag).

        
        if fromdata is None:
            raise ValueError('you must specify if `given` is data or fit result')
        fromdata = bool(fromdata)
        raw = bool(raw)
        if keepcorr is None:
            keepcorr = not raw
        if keepcorr and raw:
            raise ValueError('both keepcorr=True and raw=True')
        
        strip = False
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = [key]
            strip = True
        outslices = self._slices(outkeys)
                
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = self._concatenate(ylist)
        if y.dtype == object:
            if ycovblocks is not None:
                raise ValueError('given may contain gvars but a separate covariance matrix has been provided')
        
        self._checkpos_keys(inkeys + outkeys)
        
        Kxxs = self._assemblecovblocks(inkeys, outkeys)
        
        # if isinstance(ycovblocks, _linalg.Decomposition): # woodbury, currently un-implemented
        #     ycov = ycovblocks
        # elif ...
        if ycovblocks is not None:
            ycov = jnp.block(ycovblocks)
        elif (fromdata or raw or not keepcorr) and y.dtype == object:
            ycov = gvar.evalcov(gvar.gvar(y))
            # TODO use evalcov_blocks
            # TODO I think this ignores the case in which we are using gvars
            # and they are correlated with the GP. I guess the correct thing
            # would be to sum the data gvars to the prior ones and use the
            # resulting covariance matrix, and write a note about possible
            # different results in this case when switching raw or keepcorr.
        else:
            ycov = None
        self._check_ycov(ycov)
        
        if raw or not keepcorr or self._checkfinite:
            if y.dtype == object:
                ymean = gvar.mean(y)
            else:
                ymean = y
            self._check_ymean(ymean)
        
        if raw or not keepcorr:
            
            Kxsxs = self._assemblecovblocks(outkeys)
            
            if fromdata:
                solver = self._solver(inkeys, ycov)
            else:
                solver = self._solver(inkeys)

            mean = solver.pinv_bilinear(Kxxs, ymean)
            cov = Kxsxs - solver.ginv_quad(Kxxs)
            
            if not fromdata:
                # cov = Kxsxs - Kxsx Kxx^-1 (Kxx - ycov) Kxx^-1 Kxxs =
                #     = Kxsxs - Kxsx Kxx^-1 Kxxs + Kxsx Kxx^-1 ycov Kxx^-1 Kxxs
                if ycov is not None:
                    # if isinstance(ycov, _linalg.Decomposition): # for woodbury, currently un-implemented
                    #     ycov = ycov.matrix()
                    A = solver.ginv_linear(Kxxs)
                    # TODO do I need K⁺ here or is K⁻ fine?
                    cov += A.T @ ycov @ A
            
        else: # (keepcorr and not raw)        
            yplist = [numpy.reshape(self._prior(key), -1) for key in inkeys]
            ysplist = [numpy.reshape(self._prior(key), -1) for key in outkeys]
            yp = self._concatenate(yplist)
            ysp = self._concatenate(ysplist)
            
            if y.dtype != object and ycov is not None:
                # if isinstance(ycov, _linalg.Decomposition): # for woodbury, currently un-implemented
                #     ycov = ycov.matrix()
                y = gvar.gvar(y, ycov)
            else:
                y = numpy.asarray(y) # because y - yp fails if y is a jax array
            mat = ycov if fromdata else None
            flatout = ysp + self._solver(inkeys, mat).pinv_bilinear_robj(Kxxs, y - yp)
        
        if raw and not strip:
            meandict = {
                key: mean[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            }
            
            covdict = {
                (row, col):
                cov[rowslice, colslice].reshape(self._elements[row].shape + self._elements[col].shape)
                for row, rowslice in zip(outkeys, outslices)
                for col, colslice in zip(outkeys, outslices)
            }
            
            return meandict, covdict
            
        elif raw:
            outkey, = outkeys
            mean = mean.reshape(self._elements[outkey].shape)
            cov = cov.reshape(2 * self._elements[outkey].shape)
            return mean, cov
        
        elif not keepcorr:
                        
            flatout = gvar.gvar(mean, cov, fast=True)
        
        if not strip:
            return {
                key: flatout[slic].reshape(self._elements[key].shape)
                for key, slic in zip(outkeys, outslices)
            }
        else:
            outkey, = outkeys
            return flatout.reshape(self._elements[outkey].shape)
        
    def predfromfit(self, *args, **kw):
        """
        Like `pred` with ``fromdata=False``.
        """
        return self.pred(*args, fromdata=False, **kw)
    
    def predfromdata(self, *args, **kw):
        """
        Like `pred` with ``fromdata=True``.
        """
        return self.pred(*args, fromdata=True, **kw)
    
    def _prior_decomp(self, given, givencov=None, **kw):
        """ Internal implementation of marginal_likelihood. Keyword arguments
        are passed to _solver. """
        ylist, inkeys, ycovblocks = self._flatgiven(given, givencov)
        y = self._concatenate(ylist)
        
        self._checkpos_keys(inkeys)
        
        # Get mean.
        if y.dtype == object:
            ymean = gvar.mean(y)
        else:
            ymean = y
        self._check_ymean(ymean)
        
        # Get covariance matrix.
        # if isinstance(ycovblocks, _linalg.Decomposition): # for woodbury, currently un-implemented
        #     ycov = ycovblocks
        # elif ...
        if ycovblocks is not None:
            ycov = jnp.block(ycovblocks)
            if y.dtype == object:
                warnings.warn(f'covariance matrix may have been specified both explicitly and with gvars; the explicit one will be used')
        elif y.dtype == object:
            gvary = gvar.gvar(y)
            ycov = gvar.evalcov(gvary)
        else:
            ycov = None
        self._check_ycov(ycov)
        
        decomp = self._solver(inkeys, ycov, **kw)
        return decomp, ymean
    
    def _check_ymean(self, ymean):
        with _jaxext.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(ymean)):
                raise ValueError('mean of `given` is not finite')
    
    def _check_ycov(self, ycov):
        if ycov is None or isinstance(ycov, _linalg.Decomposition):
            return
        with _jaxext.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(ycov)):
                raise ValueError('covariance matrix of `given` is not finite')
            if self._checksym and not jnp.allclose(ycov, ycov.T):
                raise ValueError('covariance matrix of `given` is not symmetric')

    def marginal_likelihood(self, given, givencov=None, **kw):
        """
        
        Compute the logarithm of the probability of the data.
        
        The probability is computed under the Gaussian prior and Gaussian error
        model. It is also called marginal likelihood. If :math:`y` is the data
        and :math:`g` is the Gaussian process, this is
        
        .. math::
            \\log \\int p(y|g) p(g) \\mathrm{d} g.
        
        Unlike `pred`, you can't compute this with a fit result instead of
        data. If you used the Gaussian process as latent variable in a fit,
        use the whole fit to compute the marginal likelihood. E.g. `lsqfit`
        always computes the logGBF (it's the same thing).
        
        The input is an array or dictionary of arrays, ``given``. The contents
        of ``given`` represent the input data.
                
        Parameters
        ----------
        given : dictionary of arrays
            The data for some/all of the points in the GP. The arrays can
            contain either gvars or normal numbers, the latter being
            equivalent to zero-uncertainty gvars.
        givencov : dictionary of arrays, optional
            Covariance matrix of ``given``. If not specified, the covariance
            is extracted from ``given`` with ``gvar.evalcov(given)``.
        **kw :
            Additional keyword arguments are passed to the matrix decomposition.
        
        Returns
        -------
        logp : scalar
            The logarithm of the marginal likelihood.
        """
        decomp, ymean = self._prior_decomp(given, givencov, **kw)
        mll, _, _, _, _ = decomp.minus_log_normal_density(ymean, value=True)
        return -mll
    
    @staticmethod
    def _getdecomp(solver):
        return {
            'chol': _linalg.Chol,
        }[solver]
    
    @classmethod
    def decompose(cls, posdefmatrix, solver='chol', **kw):
        """
        Decompose a nonnegative definite matrix.
        
        The decomposition can be used to calculate linear algebra expressions
        where the (pseudo)inverse of the matrix appears.
        
        Parameters
        ----------
        posdefmatrix : array
            A nonnegative definite nonempty symmetric square matrix. If the
            array is not square, it must have a shape of the kind (k, n, m,
            ..., k, n, m, ...) and is reshaped to (k * n * m * ..., k * n * m *
            ...).
        solver : str
            Algorithm used to decompose the matrix.

            'chol'
                Cholesky decomposition after regularizing the matrix with a
                Gershgorin estimate of the maximum eigenvalue.
        **kw :
            Additional options.

            epsrel, epsabs : positive float or 'auto'
                Specify the threshold for considering small the eigenvalues:
                
                    eps = epsrel * maximum_eigenvalue + epsabs
                
                epsrel='auto' sets epsrel = matrix_size * float_epsilon, 
                while epsabs='auto' sets epsabs = float_epsilon. Default is
                epsrel='auto', epsabs=0.
        
        Returns
        -------
        decomp : Decomposition
            An object representing the decomposition of the matrix. The
            available methods and properties are (K being the matrix):
        
            matrix():
                Return K.
            ginv():
                Compute K⁻.
            ginv_linear(X):
                Compute K⁻X.
            pinv_bilinear(A, r)
                Compute A'K⁺r.
            pinv_bilinear_robj(A, r)
                Compute A'K⁺r, and r can be an array of arbitrary objects.
            ginv_quad(A)
                Compute A'K⁻A.
            ginv_diagquad(A)
                Compute diag(A'K⁻A).
            correlate(x)
                Compute Zx such that K = ZZ', Z can be rectangular.
            back_correlate(X)
                Compute Z'X.
            pinv_correlate(x):
                Compute Z⁺x.
            minus_log_normal_density(r, ...)
                Compute a Normal density and its derivatives.
            eps
                The threshold below which eigenvalues are not calculable.
            n
                Number of rows/columns of K.
            m
                Number of columns of Z.
        
        Notes
        -----
        The decomposition operations are JAX-traceable, but they are not meant
        to be differentiated. The method `minus_log_normal_density` provides
        required derivatives with a custom implementation, given the derivatives
        of the inputs.
        
        """
        m = jnp.asarray(posdefmatrix)
        assert m.size > 0
        assert m.ndim % 2 == 0
        half = m.ndim // 2
        head = m.shape[:half]
        tail = m.shape[half:]
        assert head == tail
        n = math.prod(head)
        m = m.reshape(n, n)
        decompcls = cls._getdecomp(solver)
        return decompcls(m, **kw)
        
        # TODO extend the interface to use composite decompositions
        # TODO accept a dict for covariance matrix



# lsqfitgp/_linalg/_pytree.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import numpy
from jax import numpy as jnp
from jax import tree_util

class AutoPyTree:
    """
    Class adding automatic recursive support for jax pytree flattening
    """
    
    def _jax_vars(self):
        """ Returns list of object attribute names which are to be considered
        children of the PyTree node """
        return [
            n for n, v in vars(self).items()
            if isinstance(v, (jnp.ndarray, numpy.ndarray, AutoPyTree))
        ]
    
    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        tree_util.register_pytree_node_class(cls)
    
    # Since I decide dinamically which members are children based on their type,
    # I have to cache the jax pytree structure aux_data such that the
    # structure is preserved when constructing an object with tree_unflatten
    # with dummies as children. This happens in jax.jacfwd for some reason.
    @functools.cached_property
    def _aux_data(self):
        jax_vars = self._jax_vars()
        other_vars = [
            (n, v) for n, v in vars(self).items()
            if n not in jax_vars
        ]
        # assert jax_vars
        return jax_vars, other_vars
    
    def tree_flatten(self):
        """JAX PyTree encoder. See `jax.tree_util.tree_flatten`."""
        jax_vars, _ = self._aux_data
        # print(f'unpacking {jax_vars} from {self.__class__.__name__}')
        children = tuple(getattr(self, n) for n in jax_vars)
        return children, self._aux_data
    
    @classmethod
    def tree_unflatten(cls, aux_data, children):
        """JAX PyTree decoder. See `jax.tree_util.tree_unflatten`."""
        self = cls.__new__(cls)
        self._aux_data = aux_data
        jax_vars, other_vars = aux_data
        for n, v in zip(jax_vars, children):
            setattr(self, n, v)
        for n, v in other_vars:
            setattr(self, n, v)
        return self



# lsqfitgp/_patch_jax.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" modifications to the global state of jax """

from jax import config
from jax import tree_util
import gvar
import numpy

config.update("jax_enable_x64", True)

class BufferDictPyTreeDef:

    @staticmethod
    def _skeleton(bd):
        """ Return a memoryless BufferDict with the same layout as `bd` """
        return gvar.BufferDict(bd, buf=numpy.empty(bd.buf.shape, []))
        # BufferDict mirrors the data type of the _buf attribute, so we do not
        # need to preserve it to maintain consistency. buf is not copied.

    def __init__(self, bd):
        self.skeleton = self._skeleton(bd)
        self.layout = {k: tuple(bd.slice_shape(k)) for k in bd.keys()}
        # it is not necessary to save the data type because that's in buf

    def __eq__(self, other):
        if not isinstance(other, __class__):
            return NotImplemented
        return self.layout == other.layout

    def __hash__(self):
        return hash(self.layout)

    def __repr__(self):
        return repr(self.layout)

    @classmethod
    def flatten(cls, bd):
        return (bd.buf,), cls(bd)

    @classmethod
    def unflatten(cls, self, children):
        buf, = children
        new = cls._skeleton(self.skeleton)
        # copy the skeleton to permit multiple unflattening
        new._extension = {}
        new._buf = buf
        return new

# register BufferDict as a pytree
tree_util.register_pytree_node(gvar.BufferDict, BufferDictPyTreeDef.flatten, BufferDictPyTreeDef.unflatten)

# TODO the current implementation of BufferDict as pytree is not really
# consistent with how JAX handles trees, because JAX expects to be allowed to
# put arbitrary objects in the leaves; in particular, internally it sometimes
# creates dummy trees filled with None. Maybe the current impl is fine with
# this; _buf gets set to None, and assuming the BufferDict is never really
# used in that crooked state, everything goes fine. The thing that this breaks
# is a subsequent flattening of the dummy, I think JAX never does this. (The
# reason for switching to buf-as-leaf in place of dict-values-as-leaves is that
# the latter breaks tracing.)
#
# Maybe since BufferDict is simple and stable, I could read its code, bypass its
# initialization altogether and set all the internal attributes to make it a
# proper pytree but also compatible with tracing.

# TODO try to drop BufferDict altogether. Currently I use it only in bcf and
# bart to pass stuff to a precompiled function. In empbayes_fit it is rebuilt
# by custom code.



# lsqfitgp/_gvarext/_format.py
#
# Copyright (c) 2023, 2024, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import math
import re
import contextlib

import gvar

def exponent(x):
    return int(math.floor(math.log10(abs(x))))

def int_mantissa(x, n, e):
    return round(x * 10 ** (n - 1 - e))

def naive_ndigits(x, n):
    log10x = math.log10(abs(x))
    n_int = int(math.floor(n))
    n_frac = n - n_int
    log10x_frac = log10x - math.floor(log10x)
    return n_int + (log10x_frac < n_frac)

def ndigits(x, n):
    ndig = naive_ndigits(x, n)
    xexp = exponent(x)
    rounded_x = int_mantissa(x, ndig, xexp) * 10 ** xexp
    if rounded_x > x:
        rounded_ndig = naive_ndigits(rounded_x, n)
        if rounded_ndig > ndig:
            x = rounded_x
            ndig = rounded_ndig
    return x, ndig

def mantissa(x, n, e):
    m = int_mantissa(x, n, e)
    s = str(abs(int(m)))
    assert len(s) == n or len(s) == n + 1 or (m == 0 and n < 0)
    if n >= 1 and len(s) == n + 1:
        e = e + 1
        s = s[:-1]
    return s, e

def insert_dot(s, n, e, *, add_leading_zeros=True, trailing_zero_char='0'):
    e = e + len(s) - n
    n = len(s)
    if e >= n - 1:
        s = s + trailing_zero_char * (e - n + 1)
    elif e >= 0:
        s = s[:1 + e] + '.' + s[1 + e:]
    elif e <= -1 and add_leading_zeros:
        s = '0' * -e + s
        s = s[:1] + '.' + s[1:]
    return s

def tostring(x):
    return '0' if x == 0 else f'{x:#.6g}'

def uformat(mu, s, errdig=2, sep=None, *,
    shareexp=True,
    outersign=False,
    uniexp=False,
    minnegexp=6,
    minposexp=4,
    padzero=None,
    possign=False,
    ):
    """
    Format a number with uncertainty.
    
    Parameters
    ----------
    mu : number
        The central value.
    s : number
        The error.
    errdig : number
        The number of digits of the error to be shown. Must be >= 1. It can be
        a noninteger, in which case the number of digits switches between the
        lower nearest integer to the upper nearest integer as the first decimal
        digit (after rounding) crosses 10 raised to the fractional part of
        `errdig`. Default 1.5.
    sep : None or str
        The separator put between the central value and the error. Eventual
        spaces must be included. If None, put the error between parentheses,
        sharing decimal places/exponential notation with the central value.
        Default None.
    shareexp : bool, default True
        Applies if `sep` is not ``None``. When using exponential notation,
        whether to share the exponent between central value and error with outer
        parentheses.
    outersign : bool
        Applied when sep is not None and shareexp is True. Whether to put the
        sign outside or within the parentheses. Default False
    uniexp : bool
        When using exponential notation, whether to use unicode characters
        instead of the standard ASCII notation. Default False.
    minnegexp : int
        The number of places after the comma at which the notation switches
        to exponential notation. Default 4. The number of places from the
        greater between central value and error is considered.
    minposexp : int
        The power of ten of the least significant digit at which exponential
        notation is used. Default 0. Setting higher values may force padding
        the error with zeros, depending on `errdig`.
    padzero : str, optional
        If provided, a character representing 0 to pad with when not using
        exponential notation due to `minposexp` even if the least significant
        digit is not on the units, instead of showing more actual digits than
        those specified.
    possign : bool, default False
        Whether to put a `+` before the central value when it is positive.

    Returns
    -------
    r : str
        The quantity (mu +/- s) nicely formatted.
    """
    if errdig < 1:
        raise ValueError('errdig < 1')
    if not math.isfinite(mu) or not math.isfinite(s) or s <= 0:
        if sep is None:
            return f'{tostring(mu)}({tostring(s)})'
        else:
            return f'{tostring(mu)}{sep}{tostring(s)}'
    
    s, sndig = ndigits(s, errdig)
    sexp = exponent(s)
    muexp = exponent(mu) if mu != 0 else sexp - sndig - 1
    smant, sexp = mantissa(s, sndig, sexp)
    mundig = sndig + muexp - sexp
    mumant, muexp = mantissa(mu, mundig, muexp)
    musign = '-' if mu < 0 else '+' if possign else ''
    
    if mundig >= sndig:
        use_exp = muexp >= mundig + minposexp or muexp <= -minnegexp
        base_exp = muexp
    else:
        use_exp = sexp >= sndig + minposexp or sexp <= -minnegexp
        base_exp = sexp
    
    if use_exp:
        mumant = insert_dot(mumant, mundig, muexp - base_exp)
        smant = insert_dot(smant, sndig, sexp - base_exp, add_leading_zeros=sep is not None)
    elif base_exp >= max(mundig, sndig) and padzero is None:
        mumant = str(abs(round(mu)))
        smant = str(abs(round(s)))
    else:
        zerochar = '0' if padzero is None else padzero
        mumant = insert_dot(mumant, mundig, muexp, trailing_zero_char=zerochar)
        if len(mumant) >= 2 and mumant.startswith('0') and all(c == zerochar for c in mumant[1:]):
            mumant = zerochar + mumant[1:]
        smant = insert_dot(smant, sndig, sexp, add_leading_zeros=sep is not None, trailing_zero_char=zerochar)
    
    if not outersign:
        mumant = musign + mumant
    
    if use_exp:
        if uniexp:
            asc = '0123456789+-'
            uni = '⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻'
            table = str.maketrans(asc, uni)
            exp = str(base_exp).translate(table)
            suffix = '×10' + exp
        else:
            suffix = f'e{base_exp:+}'
        if sep is None:
            r = mumant + '(' + smant + ')' + suffix
        elif shareexp:
            r = '(' + mumant + sep + smant + ')' + suffix
        else:
            r = mumant + suffix + sep + smant + suffix
    elif sep is None:
        r = mumant + '(' + smant + ')'
    else:
        r = mumant + sep + smant
    
    if outersign:
        r = musign + r
    
    return r

def fmtspec_kwargs(spec):
    """
    Parse a string formatting pattern to be used with `uformat`.

    Parameters
    ----------
    spec : str
        The format specification. It must follow the format

            [options](error digits)[:minimum exponent](mode)

        where brackets indicate optional parts.

    Returns
    -------
    kwargs : dict
        The keyword arguments to be passed to `uformat`.

    Notes
    -----
    Full format:

    Options: any combination these characters:
    
    '+' :
        Put a '+' before positive central values.
    '-' :
        Put the sign outside the parentheses used to group the central value and
        error mantissas in exponential notation.
    '#' :
        Do not show non-significative digits at all costs, replacing them with
        lowercase 'o', representing a rounding 0 rather than a significative 0.
    '$' :
        In exponential notation, repeat the exponent for the central value and
        error.

    Error digits: a decimal number expressing the number of leading error digits
    to show. Non-integer values indicate that the number of digits switches from
    the floor to the ceil at some value of the mantissa.

    Minimum exponent: a decimal number expressing the minimum exponent at which
    exponential notation is used.

    Mode: one of these characters:
    
    'p' :
        Put the error between parentheses.
    's' :
        Separate the central value from the error with '+/-'.
    'u' :
        Separate the central value from the error with '±'.
    'U' :
        Separate the central value from the error with '±', and use unicode
        superscript characters for exponential notation.
    """
    pat = r'([-+#$]*)(\d*\.?\d*)(:\d+)?(p|s|u|U)'
    m = re.fullmatch(pat, spec)
    if not m:
        raise ValueError(f'format specification {spec!r} not understood, format is r"{pat}"')
    kw = {}
    options = m.group(1)
    kw['possign'] = '+' in options
    kw['outersign'] = '-' in options
    kw['padzero'] = 'o' if '#' in options else None
    kw['shareexp'] = '$' not in options
    if m.group(2):
        kw['errdig'] = float(m.group(2))
    else:
        kw['errdig'] = 1.5
    if m.group(3):
        nexp = int(m.group(3)[1:])
    else:
        nexp = 5
    kw['minposexp'] = max(0, nexp - math.floor(kw['errdig']))
    kw['minnegexp'] = nexp
    mode = m.group(4)
    kw['sep'] = dict(p=None, s=' +/- ', u=' ± ', U=' ± ')[mode]
    kw['uniexp'] = mode == 'U'
    return kw

def gvar_formatter(g, spec):
    """
    A formatter for `gvar.GVar.set` that uses `uformat`.
    """
    mu = gvar.mean(g)
    s = gvar.sdev(g)
    kw = fmtspec_kwargs(spec)
    return uformat(mu, s, **kw)

@contextlib.contextmanager
def gvar_format(spec=None, *, lsqfitgp_format=True):
    """
    Context manager to set the default format specification of gvars.
    
    Parameters
    ----------
    spec : str, optional
        The format specification. If not specified and `lsqfitgp_format` is
        ``True``, use ``'#1.5p'``.
    lsqfitgp_format : bool, default True
        Whether to use a modified version of the `gvar` formatting
        specification, provided by `lsqfitgp`.

    Notes
    -----
    See `fmtspec_kwargs` for the format specification, and `uformat` for all
    details.

    See also
    --------
    gvar.fmt, gvar.GVar.set
    """
    if lsqfitgp_format:
        if spec is None:
            spec = '#1.5p'
        def formatter(g, spec, defaultspec=spec):
            if spec == '':
                spec = defaultspec
            return gvar_formatter(g, spec)
        kw = dict(formatter=formatter)
    else:
        kw = {} if spec is None else dict(default_format=spec)
    try:
        old_settings = gvar.GVar.set(**kw)
        yield
    finally:
        gvar.GVar.set(**old_settings)



# lsqfitgp/copula/_distr.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" define Distr and distribution """

import abc
import functools
import collections
import numbers
import inspect
import types
import math

import gvar
import numpy
import jax
from jax import numpy as jnp

from .. import _gvarext
from .. import _array
from .. import _signature
from . import _base

######### The following 5 functions are adapted from numpy.lib.mixins #########

def _disables_array_ufunc(obj):
    """True when __array_ufunc__ is set to None."""
    return getattr(obj, '__array_ufunc__', NotImplemented) is None

def _binary_method(ufunc, name):
    """Implement a forward binary method with a ufunc, e.g., __add__."""
    def func(self, other):
        if _disables_array_ufunc(other):
            return NotImplemented
        return ufunc(self, other)
    func.__name__ = '__{}__'.format(name)
    return func

def _reflected_binary_method(ufunc, name):
    """Implement a reflected binary method with a ufunc, e.g., __radd__."""
    def func(self, other):
        if _disables_array_ufunc(other):
            return NotImplemented
        return ufunc(other, self)
    func.__name__ = '__r{}__'.format(name)
    return func

def _numeric_methods(ufunc, name):
    """Implement forward and reflected binary methods with a ufunc."""
    return (_binary_method(ufunc, name),
            _reflected_binary_method(ufunc, name))

def _unary_method(ufunc, name):
    """Implement a unary special method with a ufunc."""
    def func(self):
        return ufunc(self)
    func.__name__ = '__{}__'.format(name)
    return func

###############################################################################

class Distr(_base.DistrBase):
    r"""

    Abstract base class to represent probability distributions.

    A `Distr` object represents a probability distribution of a variable in
    :math:`\mathbb R^n`, and provides a transformation function from a
    (multivariate) Normal variable to the target random variable.

    The main functionality is defined in `DistrBase`. The additional attributes
    and methods `params`, `signature`, and `invfcn` are not intended for common
    usage.

    Parameters
    ----------
    *params : tuple of scalar, array or Distr
        The parameters of the distribution. If the parameters have leading axes
        other than those required, the distribution is repeated i.i.d.
        over those axes. If a parameter is an instance of `Distr` itself, it
        is a random parameter and its distribution is accounted for.
    shape : int or tuple of int
        The shape of the array of i.i.d. variables to be represented, scalar by
        default. If the variable is multivariate, this shape adds as leading
        axes in the array. This shape broadcasts with the non-core shapes of the
        parameters.
    name : str, optional
        If specified, the distribution is defined for usage with
        `gvar.BufferDict` using `gvar.BufferDict.add_distribution`, and for
        convenience the constructor returns an array of gvars with the
        appropriate shape instead of the `Distr` object. See `add_distribution`.

    Returns
    -------
    If `name` is None (default):

    distr : Distr
        An object representing the distribution.

    Else:

    gvars : array of gvars
        An array of primary gvars that can be set as value in a
        `gvar.BufferDict` under a key that uses the just defined name.

    Attributes
    ----------
    params : tuple
        The parameters as passed to the constructor.
    signature : Signature
        An object representing the signature of `invfcn`. This is a class
        attribute.

    Methods
    -------
    invfcn : classmethod
        Transformation function from a (multivariate) Normal variable to the
        target random variable.

    Examples
    --------

    Use directly with `gvar.BufferDict` by setting `name`:

    >>> copula = gvar.BufferDict({
    ...     'A(x)': lgp.copula.beta(1, 1, name='A'),
    ...     'B(y)': lgp.copula.beta(3, 5, name='B'),
    ... })
    >>> copula['x']
    0.50(40)
    >>> copula['y']
    0.36(18)

    Corresponding "unrolled" usage:

    >>> A = lgp.copula.beta(1, 1)
    >>> B = lgp.copula.beta(3, 5)
    >>> A.add_distribution('A')
    >>> B.add_distribution('B')
    >>> copula = gvar.BufferDict({
    ...     'A(x)': A.gvars(),
    ...     'B(y)': B.gvars(),
    ... })

    Notice that, although the name used for `add_distribution` must be globally
    unique, for convenience it is permitted to redefine the same distribution
    family with the same parameters, even from another `Distr` instance.

    To generate automatically sensible names and avoid repeating them twice, use
    `makedict`:

    >>> lgp.copula.makedict({
    ...     'x': lgp.copula.beta(1, 1),
    ...     'y': lgp.copula.beta(3, 5),
    ... })
    BufferDict({'__copula_beta{1, 1}(x)': 0.0(1.0), '__copula_beta{3, 5}(y)': 0.0(1.0)})

    Define a distribution with a random parameter:

    >>> X = lgp.copula.halfnorm(np.sqrt(lgp.copula.invgamma(1, 1)))
    >>> X
    halfnorm(sqrt(invgamma(1, 1)))

    Now `X` represents the model

    .. math::
        \sigma^2 &\sim \mathrm{InvGamma}(1, 1), \\
        X \mid \sigma &\sim \mathrm{HalfNorm}(\sigma).

    In general it is possible to transform a `Distr` with `numpy` ufuncs and
    continuous arithmetic operations.

    Repeated usage of `Distr` instances for random parameters will share
    those parameters in the distributions. The following code:

    >>> sigma2 = lgp.copula.invgamma(1, 1)
    >>> X = lgp.copula.halfnorm(np.sqrt(sigma2))
    >>> Y = lgp.copula.halfcauchy(np.sqrt(sigma2))

    Corresponds to the model

    .. math::
        \sigma^2 &\sim \mathrm{InvGamma}(1, 1), \\
        X \mid \sigma &\sim \mathrm{HalfNorm}(\sigma), \\
        Y \mid \sigma &\sim \mathrm{HalfCauchy}(\sigma),

    with the same parameter :math:`\sigma^2` shared between the two
    distributions. However, if the distributions are now put into a
    `gvar.BufferDict`, with
    
    >>> sigma2.add_distribution('distr_sigma2')
    >>> X.add_distribution('distr_X')
    >>> Y.add_distribution('distr_Y')
    >>> bd = gvar.BufferDict({
    ...     'distr_sigma2(sigma2)': sigma2.gvars(),
    ...     'distr_X(X)': X.gvars(),
    ...     'distr_Y(Y)': Y.gvars(),
    ... })

    then this relationship breaks down; the model represented by the dictionary
    `bd` is

    .. math::
        \sigma^2 &\sim \mathrm{InvGamma}(1, 1), \\
        X \mid \sigma_X &\sim \mathrm{HalfNorm}(\sigma_X), \quad
            & \sigma_X^2 &\sim \mathrm{InvGamma}(1, 1), \\
        Y \mid \sigma_Y &\sim \mathrm{HalfCauchy}(\sigma_Y), \quad
            & \sigma_Y^2 &\sim \mathrm{InvGamma}(1, 1),

    with separate, independent parameters :math:`\sigma,\sigma_X,\sigma_Y`,
    because each dictionary entry is evaluated separately. Indeed, trying to do
    this with `makedict` will raise an error:

    >>> bd = lgp.copula.makedict({'sigma2': sigma2, 'X': X, 'Y': Y})
    ValueError: cross-key occurrences of object(s):
    invgamma with id 6201535248: <sigma2>, <X.0.0>, <Y.0.0>

    To use all the distributions at once while preserving the relationships,
    put them into a container of choice and wrap it as a `Copula` object:

    >>> sigmaXY = lgp.copula.Copula({'sigma2': sigma2, 'X': X, 'Y': Y})

    The `Copula` provides a `partial_invfcn` function to map Normal variables
    to a structure, with the same layout as the input one, of desired variates.
    The whole `Copula` can be used in `gvar.BufferDict`:

    >>> bd = lgp.copula.makedict({'sigmaXY': sigmaXY})
    >>> bd
    BufferDict({"__copula_{'sigma2': invgamma{1, 1}, 'X': halfnorm{sqrt{_Path{path=[{DictKey{key='sigma2'},}]}}}, 'Y': halfcauchy{sqrt{_Path{path=[{DictKey{key='sigma2'},}]}}}}(sigmaXY)": array([0.0(1.0), 0.0(1.0), 0.0(1.0)], dtype=object)})
    >>> bd['sigmaXY']
    {'sigma2': 1.4(1.7), 'X': 0.81(89), 'Y': 1.2(1.7)}
    >>> gvar.corr(bd['sigmaXY']['X'], bd['sigmaXY']['Y'])
    0.21950577757757836

    Although the actual dictionary value is a flat array, getting the unwrapped
    key reproduces the original structure.
    
    To apply arbitrary transformations, use manually `invfcn`:

    >>> @functools.partial(lgp.gvar_gufunc, signature='(n)->(n)')
    >>> @functools.partial(jnp.vectorize, signature='(n)->(n)')
    >>> def model_invfcn(normal_params):
    ...     sigma2 = lgp.copula.invgamma.invfcn(normal_params[0], 1, 1)
    ...     sigma = jnp.sqrt(sigma2)
    ...     X = lgp.copula.halfnorm.invfcn(normal_params[1], sigma)
    ...     Y = lgp.copula.halfcauchy.invfcn(normal_params[2], sigma)
    ...     return jnp.stack([sigma, X, Y])
    
    The `jax.numpy.vectorize` decorator makes `model_invfcn` support
    broadcasting on additional input axes, while `gvar_gufunc` makes it accept
    gvars as input.

    See also
    --------
    DistrBase, Copula, gvar.BufferDict.uniform

    Notes
    -----
    Concrete subclasses must define `invfcn`, and define the class attribute
    `signature` to the numpy signature string of `invfcn`, unless `invfcn` is an
    ufunc and its number of parameters can be inferred. `invfcn` must be
    vectorized.

    """

    @classmethod
    @abc.abstractmethod
    def invfcn(cls, x, *params):
        r"""

        Normal to desired distribution transformation.

        Maps a (multivariate) Normal variable to a variable with the desired
        marginal distribution. In symbols: :math:`y = F^{-1}(\Phi(x))`. This
        function is a generalized ufunc, jax traceable, vmappable one time, and
        differentiable one time. The signature is accessible through the
        class attribute `signature`.

        Parameters
        ----------
        x : array_like
            The input Normal variable.
        *params : array_like
            The parameters of the distribution.

        Returns
        -------
        y : array_like
            The output variable with the desired marginal distribution.

        """
        pass

    def _get_x_core_shape(self, *preprocessed_params):
        sig = self.signature.eval(None, *preprocessed_params)
        return sig.core_in_shapes[0]

    def _eval_shapes(self, shape):

        # check number of parameters
        if self.signature.nin != 1 + len(self.params):
            raise TypeError(f'{self.__class__.__name__} distribution has '
                f'{self.signature.nin - 1} parameters, but {len(self.params)} '
                'parameters were passed to the constructor')

        # convert shape to tuple
        if isinstance(shape, numbers.Integral):
            shape = (shape,)
        else:
            shape = tuple(shape)

        # make sure parameters have a shape
        array_params = [
            p if hasattr(p, 'shape') else jnp.asarray(p)
            for p in self.params
        ]

        # parse signature of cls.invfcn
        x_core_shape = self._get_x_core_shape(*array_params)
        x = jax.ShapeDtypeStruct(shape + x_core_shape, 'd')
        sig = self.signature.eval(x, *array_params)
        self._in_shape_1 = sig.in_shapes[0]
        self.distrshape, = sig.core_out_shapes
        self.shape, = sig.out_shapes
        
        self._compute_in_shape()

    def _compute_in_shape(self):
        in_size = math.prod(self._in_shape_1)
        cache = set()
        for p in self.params:
            if isinstance(p, __class__):
                in_size += p._compute_in_size(cache)
        if in_size == 1:
            self.in_shape = ()
        else:
            self.in_shape = in_size,
        self._ancestor_count = len(cache)

    def _compute_in_size(self, cache):
        if (out := super()._compute_in_size(cache)) is not None:
            return out
        in_size = math.prod(self._in_shape_1)
        for p in self.params:
            if isinstance(p, __class__):
                in_size += p._compute_in_size(cache)
        return in_size

    def _partial_invfcn_internal(self, x, i, cache):
        if (out := super()._partial_invfcn_internal(x, i, cache)) is not None:
            return out

        concrete_params = []
        for p in self.params:
            
            if isinstance(p, __class__):
                p, i = p._partial_invfcn_internal(x, i, cache)
            else:
                p = jnp.asarray(p)
            
            concrete_params.append(p)

        in_size = math.prod(self._in_shape_1)
        assert i + in_size <= x.size
        last = x[i:i + in_size].reshape(self._in_shape_1)
        
        y = self.invfcn(last, *concrete_params)
        if y.shape != self.shape or y.dtype != self.dtype:
            raise ValueError(f'{self.__class__.__name__}.invfcn returned '
                f'array with shape {y.shape} and dtype {y.dtype}, while '
                f'{self.shape} and {self.dtype} were expected')
        
        cache[self] = y
        return y, i + in_size

    @functools.cached_property
    def _partial_invfcn(self):

        # determine signature
        shapestr = lambda shape: ','.join(map(str, shape))
        signature = f'({shapestr(self.in_shape)})->({shapestr(self.shape)})'

        # wrap to support gvars
        @functools.partial(_gvarext.gvar_gufunc, signature=signature)
        # @jax.jit
        @functools.partial(jnp.vectorize, signature=signature)
        def _partial_invfcn(x):
            assert x.shape == self.in_shape
            if not self.in_shape:
                x = x[None]
            cache = {}
            y, i = self._partial_invfcn_internal(x, 0, cache)
            assert i == x.size
            assert len(cache) == 1 + self._ancestor_count
            return y

        return _partial_invfcn

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)

        # check and/or set signature attribute (the gufunc signature of invfcn)
        if not hasattr(cls, 'signature'):
            sig = inspect.signature(cls.invfcn)
            if not all(
                p.kind in (inspect.Parameter.POSITIONAL_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                for p in sig.parameters.values()
            ):
                raise ValueError('can not automatically infer signature of '
                    f'{cls.__qualname__}.invfcn')
            cls.signature = ','.join(['()'] * len(sig.parameters)) + '->()'
        if not isinstance(cls.signature, _signature.Signature):
            cls.signature = _signature.Signature(cls.signature)
        cls.signature.check_nargs(cls.invfcn)
        
        # set dtype to float if not specified
        if getattr(cls, 'dtype', NotImplemented) is NotImplemented:
            cls.dtype = jax.dtypes.canonicalize_dtype(jnp.float64)

        # set __signature__ to take positional parameters from invfcn
        sig = inspect.signature(cls.invfcn)
        pos_params = list(sig.parameters.values())[1:]
        sig = inspect.signature(cls.__new__)
        key_params = [
            p for i, p in enumerate(sig.parameters.values())
            if p.kind in (inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD)
            and i > 0
        ]
        cls.__signature__ = inspect.Signature(pos_params + key_params)

    def __new__(cls, *params, shape=(), name=None):

        self = super().__new__(cls)
        self.params = params
        self._eval_shapes(shape)

        if name is None:
            return self
        else:
            self.add_distribution(name)
            return self.gvars()
    
    class _Descr(collections.namedtuple('Distr', 'family shape params')):
        """ static representation of a Distr object """

        def __repr__(self):
            args = list(map(repr, self.params))
            if len(self.shape) == 1:
                args += [f'shape={self.shape[0]}']
            elif self.shape:
                args += [f'shape={self.shape}']
            arglist = ', '.join(args)
            return f'{self.family.__name__}({arglist})'

    def _compute_staticdescr(self, path, cache):
        if (obj := super()._compute_staticdescr(path, cache)) is not None:
            return obj
        
        params = []
        for i, p in enumerate(self.params):
            if isinstance(p, __class__):
                p = p._compute_staticdescr(path + [i], cache)
            else:
                p = numpy.asarray(p).tolist()
            params.append(p)

        return self._Descr(self.__class__, self.shape, tuple(params))

    def _shapestr(self, shape):
        if shape:
            return (str(shape)
                .replace(',)', ')')
                .replace('(' , '[')
                .replace(')' , ']')
                .replace(' ', '')
            )
        else:
            return ''
        
    def __repr__(self, path='', cache=None):

        if isinstance(cache := super().__repr__(path, cache), str):
            return cache
        
        args = []
        for i, p in enumerate(self.params):
            
            if isinstance(p, __class__):
                p = p.__repr__('.'.join((path, str(i))).lstrip('.'), cache)
            elif hasattr(p, 'shape'):
                p = f'Array{self._shapestr(p.shape)}'
            else:
                p = repr(p)
            args.append(p)

        if len(self.shape) == 1:
            args += [f'shape={self.shape[0]}']
        elif self.shape:
            args += [f'shape={self.shape}']

        return f'{self.__class__.__name__}({", ".join(args)})'

    def __array_ufunc__(self, ufunc, method, *inputs, **kw):
        if method != '__call__' or kw or ufunc.signature:
            # TODO jax 0.4.15 should introduce ufunc methods
            return NotImplemented
        ufunc_class = UFunc.make_subclass(ufunc)
        return ufunc_class(*inputs)

        # TODO make this work with gufuncs. See comment in _signature.py.
        # matmul in particular.

    # continuous binary operations
    __add__, __radd__ = _numeric_methods(numpy.add, 'add')
    __sub__, __rsub__ = _numeric_methods(numpy.subtract, 'sub')
    __mul__, __rmul__ = _numeric_methods(numpy.multiply, 'mul')
    # __matmul__, __rmatmul__ = _numeric_methods(numpy.matmul, 'matmul')
    __truediv__, __rtruediv__ = _numeric_methods(numpy.divide, 'truediv')
    __mod__, __rmod__ = _numeric_methods(numpy.remainder, 'mod')
    __divmod__, __rdivmod__ = _numeric_methods(numpy.divmod, 'divmod')
    __pow__, __rpow__ = _numeric_methods(numpy.power, 'pow')

    # continuous unary operations
    __neg__ = _unary_method(numpy.negative, 'neg')
    __pos__ = _unary_method(numpy.positive, 'pos')
    __abs__ = _unary_method(numpy.absolute, 'abs')

    # TODO add __getitem__ and __array_function__

class UFunc:
    """ base class of objects representing ufuncs applied to Distr instances """

    def __new__(cls, *args):
        return super().__new__(cls, *args)
        # this __new__ serves to forbid keyword arguments

    @classmethod
    def invfcn(cls, x, *args):
        return cls._ufunc(*args)

    def _get_x_core_shape(self, *_):
        return (0,)

    @classmethod
    @functools.lru_cache(maxsize=None) # functools.cache not available in 3.8
    def make_subclass(cls, ufunc):
        def exec_body(ns):
            ns['_ufunc'] = getattr(jnp, ufunc.__name__)
            ns['signature'] = ','.join(['(0)'] + ufunc.nin * ['()']) + '->()'
        return types.new_class(ufunc.__name__, (__class__, Distr), exec_body=exec_body)

def distribution(invfcn, signature=None, dtype=None):
    r"""

    Decorator to define a distribution from a transformation function.

    Parameters
    ----------
    invfcn : function
        The transformation function from a (multivariate) standard Normal
        variable to the target random variable. The signature must be
        ``invfcn(x, *params)``. It must be jax-traceable. It does not need to
        be vectorized.
    signature : str, optional
        The signature of `invfcn`, as a numpy signature string. If not
        specified, `invfcn` is assumed to take and output scalars.
    dtype : dtype, optional
        The dtype of the output of `invfcn`. If not specified, it is assumed to
        be floating point.

    Returns
    -------
    cls : Distr
        The new distribution class.

    Examples
    --------

    >>> @lgp.copula.distribution
    ... def uniform(x, a, b):
    ...     return a + (b - a) * jax.scipy.stats.norm.cdf(x)

    >>> @functools.partial(lgp.copula.distribution, signature='(n,m)->(n)')
    ... def wishart(x):
    ...     " this parametrization is terrible, do not use "
    ...     return x @ x.T

    """
    
    def exec_body(ns):
        if signature is not None:
            ns['signature'] = signature
        if dtype is not None:
            ns['dtype'] = dtype
        ns['invfcn'] = staticmethod(jnp.vectorize(invfcn, signature=signature))

    return types.new_class(invfcn.__name__, (Distr,), exec_body=exec_body)



# lsqfitgp/copula/_copulas.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" predefined distributions """

import functools
import collections

from jax.scipy import special as jspecial
import jax
from jax import numpy as jnp

from .. import _jaxext
from .. import _array
from . import _beta, _gamma
from . import _distr

def _normcdf(x):
    x = jnp.asarray(x)
    x = x.astype(_jaxext.float_type(x))
    return jspecial.ndtr(x)

    # In jax < 0.?.?, jax.scipy.stats.norm.sf is implemented as 1 - cdf(x)
    # instead of cdf(-x), defeating the purpose of numerical accuracy. Use
    # _normcdf(-x) instead. See https://github.com/google/jax/issues/17199

class beta(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Beta_distribution
    """
    
    @staticmethod
    def invfcn(x, alpha, beta):
        return _beta.beta.ppf(_normcdf(x), a=alpha, b=beta)

class dirichlet(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Dirichlet_distribution
    """

    signature = '(n),(n)->(n)'
    
    @classmethod
    def invfcn(cls, x, alpha):
        lny = loggamma.invfcn(x, alpha)
        norm = jspecial.logsumexp(lny, axis=-1, keepdims=True)
        return jnp.exp(lny - norm)

    # @classmethod
    # def _invfcn_tiny_alpha(cls, x, alpha):
    #     q = _normcdf(x)
    #     lnq = jnp.log(q)
    #     lny = lnq / alpha
    #     lnnorm = jspecial.logsumexp(lny, axis=-1, keepdims=True)
    #     return jnp.exp(lny - lnnorm)

        # For a -> 0:
        #
        # gamma.cdf(x, a) = P(a, x)
        #                 = gamma(a, x) / Gamma(a)
        #                 = int_0^x dt e^-t t^(a - 1) / (1 / a)
        #                 = a [t^a / a]_0^x
        #                 = a x^a / a
        #                 = x^a
        #
        # gamma.ppf(q, a) = P^-1(a, q)
        #                 = q^1/a

class gamma(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Gamma_distribution
    """
    
    @staticmethod
    def _boundary(x):
        return {
            jnp.dtype(jnp.float32): 12,
            jnp.dtype(jnp.float64): 37,
        }[x.dtype]

    @classmethod
    def invfcn(cls, x, alpha, beta):
        x = jnp.asarray(x)
        x = x.astype(_jaxext.float_type(x))
        boundary = cls._boundary(x)
        return _piecewise_multiarg(
            [x < 0, x < boundary, x >= boundary],
                # TODO the x < 0 case is probably never considered because
                # piecewise evaluates from the right and x < boundary is
                # satisfied too. Why are the tests not uncovering the
                # inaccuracy? First find whether it's accurate the same or if
                # the tests are lacking, then correct the conditionals.
            [
                lambda x, a: _gamma.gamma.ppf(_normcdf(x), a),
                lambda x, a: _gamma.gamma.isf(_normcdf(-x), a),
                lambda x, a: _gamma._gammaisf_normcdf_large_neg_x(-x, a),
            ],
            x, alpha,
        ) / beta

class loggamma(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Gamma_distribution, `scipy.stats.loggamma`

    This is the distribution of the logarithm of a Gamma variable. The naming
    convention is the opposite of lognorm, which is the distribution of the
    exponential of a Normal variable.
    """
    
    @staticmethod
    def _boundary(x):
        return gamma._boundary(x)

    @classmethod
    def invfcn(cls, x, alpha):
        x = jnp.asarray(x)
        x = x.astype(_jaxext.float_type(x))
        boundary = cls._boundary(x)
        return _piecewise_multiarg(
            [x < 0, x < boundary, x >= boundary],
            [
                lambda x, alpha: _gamma.loggamma.ppf(_normcdf(x), alpha),
                lambda x, alpha: _gamma.loggamma.isf(_normcdf(-x), alpha),
                lambda x, alpha: _gamma._loggammaisf_normcdf_large_neg_x(-x, alpha),
            ],
            x, alpha,
        )

    # TODO scipy.stats.gamma has inaccurate logsf instead of using loggamma.sf,
    # open an issue

class invgamma(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Inverse-gamma_distribution
    """
    
    @staticmethod
    def _boundary(x):
        return -gamma._boundary(x)

    @classmethod
    def invfcn(cls, x, alpha, beta):
        x = jnp.asarray(x)
        x = x.astype(_jaxext.float_type(x))
        boundary = cls._boundary(x)
        return beta * _piecewise_multiarg(
            [x < boundary, x < 0, x >= 0],
            [
                lambda x, a: 1 / _gamma._gammaisf_normcdf_large_neg_x(x, a),
                lambda x, a: _gamma.invgamma.ppf(_normcdf(x), a),
                lambda x, a: _gamma.invgamma.isf(_normcdf(-x), a),
            ],
            x, alpha,
        )

def _piecewise_multiarg(conds, functions, *operands):
    conds = jnp.stack(conds, axis=-1)
    index = jnp.argmax(conds, axis=-1)
    return _vectorized_switch(index, functions, *operands)

@functools.partial(jnp.vectorize, excluded=(1,))
def _vectorized_switch(index, branches, *operands):
    return jax.lax.switch(index, branches, *operands)

class halfcauchy(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Cauchy_distribution, `scipy.stats.halfcauchy`
    """
    
    @staticmethod
    def _ppf(p):
        return jnp.tan(jnp.pi * p / 2)
    
    @staticmethod
    def _isf(p):
        return 1 / jnp.tan(jnp.pi * p / 2)
    
    @classmethod
    def invfcn(cls, x, gamma):
        return gamma * jnp.where(x < 0,
            cls._ppf(_normcdf(x)),
            cls._isf(_normcdf(-x)),
        )

class halfnorm(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Half-normal_distribution
    """
    
    @staticmethod
    def _ppf(p):
        # F(x) = 2 Φ(x) - 1
        # -->  F⁻¹(p) = Φ⁻¹((1 + p) / 2)
        return jspecial.ndtri((1 + p) / 2)

    @staticmethod
    def _isf(p):
        # Φ(-x) = 1 - Φ(x)
        # -->  Φ⁻¹(1 - p) = -Φ⁻¹(p)
        # S(x) = 1 - F(x)
        # -->  S⁻¹(p) = F⁻¹(1 - p)
        #             = Φ⁻¹((2 - p) / 2)
        #             = Φ⁻¹(1 - p / 2)
        #             = -Φ⁻¹(p / 2)
        return -jspecial.ndtri(p / 2)

    @classmethod
    def invfcn(cls, x, sigma):
        return sigma * jnp.where(x < 0,
            cls._ppf(_normcdf(x)),
            cls._isf(_normcdf(-x)),
        )

class uniform(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Continuous_uniform_distribution
    """
    
    @staticmethod
    def invfcn(x, a, b):
        return a + (b - a) * _normcdf(x)

class lognorm(_distr.Distr):
    """
    https://en.wikipedia.org/wiki/Log-normal_distribution
    """
    
    @staticmethod
    def invfcn(x, mu, sigma):
        return jnp.exp(mu + sigma * x)
